from __future__ import annotations

from gitlab.base import RESTObject
from gitlab.mixins import CRUDMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import ArrayAttribute, RequiredOptional

__all__ = [
    "ProjectRelease",
    "ProjectReleaseManager",
    "ProjectReleaseLink",
    "ProjectReleaseLinkManager",
]


class ProjectRelease(SaveMixin, RESTObject):
    _id_attr = "tag_name"

    links: ProjectReleaseLinkManager


class ProjectReleaseManager(CRUDMixin[ProjectRelease]):
    _path = "/projects/{project_id}/releases"
    _obj_cls = ProjectRelease
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("tag_name",), optional=("name", "description", "ref", "assets")
    )
    _list_filters = ("order_by", "sort", "include_html_description")
    _update_attrs = RequiredOptional(
        optional=("name", "description", "milestones", "released_at")
    )
    _types = {"milestones": ArrayAttribute}


class ProjectReleaseLink(ObjectDeleteMixin, SaveMixin, RESTObject):
    pass


class ProjectReleaseLinkManager(CRUDMixin[ProjectReleaseLink]):
    _path = "/projects/{project_id}/releases/{tag_name}/assets/links"
    _obj_cls = ProjectReleaseLink
    _from_parent_attrs = {"project_id": "project_id", "tag_name": "tag_name"}
    _create_attrs = RequiredOptional(
        required=("name", "url"),
        optional=("filepath", "direct_asset_path", "link_type"),
    )
    _update_attrs = RequiredOptional(
        optional=("name", "url", "filepath", "direct_asset_path", "link_type")
    )



from gitlab.base import RESTObject
from gitlab.mixins import ListMixin

__all__ = ["ProjectMergeTrain", "ProjectMergeTrainManager"]


class ProjectMergeTrain(RESTObject):
    pass


class ProjectMergeTrainManager(ListMixin[ProjectMergeTrain]):
    _path = "/projects/{project_id}/merge_trains"
    _obj_cls = ProjectMergeTrain
    _from_parent_attrs = {"project_id": "id"}
    _list_filters = ("scope",)



from __future__ import annotations

from gitlab.base import RESTObject
from gitlab.mixins import ListMixin, RetrieveMixin, SaveMixin, UpdateMixin
from gitlab.types import RequiredOptional

__all__ = [
    "ProjectResourceGroup",
    "ProjectResourceGroupManager",
    "ProjectResourceGroupUpcomingJob",
    "ProjectResourceGroupUpcomingJobManager",
]


class ProjectResourceGroup(SaveMixin, RESTObject):
    _id_attr = "key"

    upcoming_jobs: ProjectResourceGroupUpcomingJobManager


class ProjectResourceGroupManager(
    RetrieveMixin[ProjectResourceGroup], UpdateMixin[ProjectResourceGroup]
):
    _path = "/projects/{project_id}/resource_groups"
    _obj_cls = ProjectResourceGroup
    _from_parent_attrs = {"project_id": "id"}
    _list_filters = ("order_by", "sort", "include_html_description")
    _update_attrs = RequiredOptional(optional=("process_mode",))


class ProjectResourceGroupUpcomingJob(RESTObject):
    pass


class ProjectResourceGroupUpcomingJobManager(
    ListMixin[ProjectResourceGroupUpcomingJob]
):
    _path = "/projects/{project_id}/resource_groups/{resource_group_key}/upcoming_jobs"
    _obj_cls = ProjectResourceGroupUpcomingJob
    _from_parent_attrs = {"project_id": "project_id", "resource_group_key": "key"}



"""
Defines http backends for processing http requests
"""

from .requests_backend import (
    JobTokenAuth,
    OAuthTokenAuth,
    PrivateTokenAuth,
    RequestsBackend,
    RequestsResponse,
)

DefaultBackend = RequestsBackend
DefaultResponse = RequestsResponse

__all__ = [
    "DefaultBackend",
    "DefaultResponse",
    "JobTokenAuth",
    "OAuthTokenAuth",
    "PrivateTokenAuth",
]



from gitlab.base import RESTObject
from gitlab.mixins import CreateMixin, DeleteMixin, ListMixin, ObjectDeleteMixin
from gitlab.types import RequiredOptional

__all__ = ["Application", "ApplicationManager"]


class Application(ObjectDeleteMixin, RESTObject):
    _url = "/applications"
    _repr_attr = "name"


class ApplicationManager(
    ListMixin[Application], CreateMixin[Application], DeleteMixin[Application]
):
    _path = "/applications"
    _obj_cls = Application
    _create_attrs = RequiredOptional(
        required=("name", "redirect_uri", "scopes"), optional=("confidential",)
    )



from __future__ import annotations

from typing import Any, BinaryIO, TYPE_CHECKING

import requests

import gitlab
from gitlab import cli
from gitlab import exceptions as exc
from gitlab import types
from gitlab.base import RESTObject, TObjCls
from gitlab.mixins import (
    CreateMixin,
    CRUDMixin,
    DeleteMixin,
    ListMixin,
    NoUpdateMixin,
    ObjectDeleteMixin,
    SaveMixin,
)
from gitlab.types import RequiredOptional

from .access_requests import GroupAccessRequestManager  # noqa: F401
from .audit_events import GroupAuditEventManager  # noqa: F401
from .badges import GroupBadgeManager  # noqa: F401
from .boards import GroupBoardManager  # noqa: F401
from .branches import GroupProtectedBranchManager  # noqa: F401
from .clusters import GroupClusterManager  # noqa: F401
from .container_registry import GroupRegistryRepositoryManager  # noqa: F401
from .custom_attributes import GroupCustomAttributeManager  # noqa: F401
from .deploy_tokens import GroupDeployTokenManager  # noqa: F401
from .epics import GroupEpicManager  # noqa: F401
from .export_import import GroupExportManager, GroupImportManager  # noqa: F401
from .group_access_tokens import GroupAccessTokenManager  # noqa: F401
from .hooks import GroupHookManager  # noqa: F401
from .invitations import GroupInvitationManager  # noqa: F401
from .issues import GroupIssueManager  # noqa: F401
from .iterations import GroupIterationManager  # noqa: F401
from .labels import GroupLabelManager  # noqa: F401
from .member_roles import GroupMemberRoleManager  # noqa: F401
from .members import (  # noqa: F401
    GroupBillableMemberManager,
    GroupMemberAllManager,
    GroupMemberManager,
)
from .merge_request_approvals import GroupApprovalRuleManager
from .merge_requests import GroupMergeRequestManager  # noqa: F401
from .milestones import GroupMilestoneManager  # noqa: F401
from .notification_settings import GroupNotificationSettingsManager  # noqa: F401
from .packages import GroupPackageManager  # noqa: F401
from .projects import GroupProjectManager, SharedProjectManager  # noqa: F401
from .push_rules import GroupPushRulesManager
from .runners import GroupRunnerManager  # noqa: F401
from .service_accounts import GroupServiceAccountManager  # noqa: F401
from .statistics import GroupIssuesStatisticsManager  # noqa: F401
from .variables import GroupVariableManager  # noqa: F401
from .wikis import GroupWikiManager  # noqa: F401

__all__ = [
    "Group",
    "GroupManager",
    "GroupDescendantGroup",
    "GroupDescendantGroupManager",
    "GroupLDAPGroupLink",
    "GroupLDAPGroupLinkManager",
    "GroupSubgroup",
    "GroupSubgroupManager",
    "GroupSAMLGroupLink",
    "GroupSAMLGroupLinkManager",
]


class Group(SaveMixin, ObjectDeleteMixin, RESTObject):
    _repr_attr = "name"

    access_tokens: GroupAccessTokenManager
    accessrequests: GroupAccessRequestManager
    approval_rules: GroupApprovalRuleManager
    audit_events: GroupAuditEventManager
    badges: GroupBadgeManager
    billable_members: GroupBillableMemberManager
    boards: GroupBoardManager
    clusters: GroupClusterManager
    customattributes: GroupCustomAttributeManager
    deploytokens: GroupDeployTokenManager
    descendant_groups: GroupDescendantGroupManager
    epics: GroupEpicManager
    exports: GroupExportManager
    hooks: GroupHookManager
    imports: GroupImportManager
    invitations: GroupInvitationManager
    issues: GroupIssueManager
    issues_statistics: GroupIssuesStatisticsManager
    iterations: GroupIterationManager
    labels: GroupLabelManager
    ldap_group_links: GroupLDAPGroupLinkManager
    member_roles: GroupMemberRoleManager
    members: GroupMemberManager
    members_all: GroupMemberAllManager
    mergerequests: GroupMergeRequestManager
    milestones: GroupMilestoneManager
    notificationsettings: GroupNotificationSettingsManager
    packages: GroupPackageManager
    projects: GroupProjectManager
    shared_projects: SharedProjectManager
    protectedbranches: GroupProtectedBranchManager
    pushrules: GroupPushRulesManager
    registry_repositories: GroupRegistryRepositoryManager
    runners: GroupRunnerManager
    subgroups: GroupSubgroupManager
    variables: GroupVariableManager
    wikis: GroupWikiManager
    saml_group_links: GroupSAMLGroupLinkManager
    service_accounts: GroupServiceAccountManager

    @cli.register_custom_action(cls_names="Group", required=("project_id",))
    @exc.on_http_error(exc.GitlabTransferProjectError)
    def transfer_project(self, project_id: int, **kwargs: Any) -> None:
        """Transfer a project to this group.

        Args:
            to_project_id: ID of the project to transfer
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabTransferProjectError: If the project could not be transferred
        """
        path = f"/groups/{self.encoded_id}/projects/{project_id}"
        self.manager.gitlab.http_post(path, **kwargs)

    @cli.register_custom_action(cls_names="Group", required=(), optional=("group_id",))
    @exc.on_http_error(exc.GitlabGroupTransferError)
    def transfer(self, group_id: int | None = None, **kwargs: Any) -> None:
        """Transfer the group to a new parent group or make it a top-level group.

        Requires GitLab ≥14.6.

        Args:
            group_id: ID of the new parent group. When not specified,
                the group to transfer is instead turned into a top-level group.
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGroupTransferError: If the group could not be transferred
        """
        path = f"/groups/{self.encoded_id}/transfer"
        post_data = {}
        if group_id is not None:
            post_data["group_id"] = group_id
        self.manager.gitlab.http_post(path, post_data=post_data, **kwargs)

    @cli.register_custom_action(cls_names="Group", required=("scope", "search"))
    @exc.on_http_error(exc.GitlabSearchError)
    def search(
        self, scope: str, search: str, **kwargs: Any
    ) -> gitlab.GitlabList | list[dict[str, Any]]:
        """Search the group resources matching the provided string.

        Args:
            scope: Scope of the search
            search: Search string
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabSearchError: If the server failed to perform the request

        Returns:
            A list of dicts describing the resources found.
        """
        data = {"scope": scope, "search": search}
        path = f"/groups/{self.encoded_id}/search"
        return self.manager.gitlab.http_list(path, query_data=data, **kwargs)

    @cli.register_custom_action(cls_names="Group")
    @exc.on_http_error(exc.GitlabCreateError)
    def ldap_sync(self, **kwargs: Any) -> None:
        """Sync LDAP groups.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabCreateError: If the server cannot perform the request
        """
        path = f"/groups/{self.encoded_id}/ldap_sync"
        self.manager.gitlab.http_post(path, **kwargs)

    @cli.register_custom_action(
        cls_names="Group",
        required=("group_id", "group_access"),
        optional=("expires_at",),
    )
    @exc.on_http_error(exc.GitlabCreateError)
    def share(
        self,
        group_id: int,
        group_access: int,
        expires_at: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Share the group with a group.

        Args:
            group_id: ID of the group.
            group_access: Access level for the group.
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabCreateError: If the server failed to perform the request

        Returns:
            Group
        """
        path = f"/groups/{self.encoded_id}/share"
        data = {
            "group_id": group_id,
            "group_access": group_access,
            "expires_at": expires_at,
        }
        server_data = self.manager.gitlab.http_post(path, post_data=data, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(server_data, dict)
        self._update_attrs(server_data)

    @cli.register_custom_action(cls_names="Group", required=("group_id",))
    @exc.on_http_error(exc.GitlabDeleteError)
    def unshare(self, group_id: int, **kwargs: Any) -> None:
        """Delete a shared group link within a group.

        Args:
            group_id: ID of the group.
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabDeleteError: If the server failed to perform the request
        """
        path = f"/groups/{self.encoded_id}/share/{group_id}"
        self.manager.gitlab.http_delete(path, **kwargs)

    @cli.register_custom_action(cls_names="Group")
    @exc.on_http_error(exc.GitlabRestoreError)
    def restore(self, **kwargs: Any) -> None:
        """Restore a  group marked for deletion..

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabRestoreError: If the server failed to perform the request
        """
        path = f"/groups/{self.encoded_id}/restore"
        self.manager.gitlab.http_post(path, **kwargs)


class GroupManager(CRUDMixin[Group]):
    _path = "/groups"
    _obj_cls = Group
    _list_filters = (
        "skip_groups",
        "all_available",
        "search",
        "order_by",
        "sort",
        "statistics",
        "owned",
        "with_custom_attributes",
        "min_access_level",
        "top_level_only",
    )
    _create_attrs = RequiredOptional(
        required=("name", "path"),
        optional=(
            "description",
            "membership_lock",
            "visibility",
            "share_with_group_lock",
            "require_two_factor_authentication",
            "two_factor_grace_period",
            "project_creation_level",
            "auto_devops_enabled",
            "subgroup_creation_level",
            "emails_disabled",
            "avatar",
            "mentions_disabled",
            "lfs_enabled",
            "request_access_enabled",
            "parent_id",
            "default_branch_protection",
            "shared_runners_minutes_limit",
            "extra_shared_runners_minutes_limit",
        ),
    )
    _update_attrs = RequiredOptional(
        optional=(
            "name",
            "path",
            "description",
            "membership_lock",
            "share_with_group_lock",
            "visibility",
            "require_two_factor_authentication",
            "two_factor_grace_period",
            "project_creation_level",
            "auto_devops_enabled",
            "subgroup_creation_level",
            "emails_disabled",
            "avatar",
            "mentions_disabled",
            "lfs_enabled",
            "request_access_enabled",
            "default_branch_protection",
            "file_template_project_id",
            "shared_runners_minutes_limit",
            "extra_shared_runners_minutes_limit",
            "prevent_forking_outside_group",
            "shared_runners_setting",
        )
    )
    _types = {"avatar": types.ImageAttribute, "skip_groups": types.ArrayAttribute}

    @exc.on_http_error(exc.GitlabImportError)
    def import_group(
        self,
        file: BinaryIO,
        path: str,
        name: str,
        parent_id: int | str | None = None,
        **kwargs: Any,
    ) -> dict[str, Any] | requests.Response:
        """Import a group from an archive file.

        Args:
            file: Data or file object containing the group
            path: The path for the new group to be imported.
            name: The name for the new group.
            parent_id: ID of a parent group that the group will
                be imported into.
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabImportError: If the server failed to perform the request

        Returns:
            A representation of the import status.
        """
        files = {"file": ("file.tar.gz", file, "application/octet-stream")}
        data: dict[str, Any] = {"path": path, "name": name}
        if parent_id is not None:
            data["parent_id"] = parent_id

        return self.gitlab.http_post(
            "/groups/import", post_data=data, files=files, **kwargs
        )


class SubgroupBaseManager(ListMixin[TObjCls]):
    _from_parent_attrs = {"group_id": "id"}
    _list_filters = (
        "skip_groups",
        "all_available",
        "search",
        "order_by",
        "sort",
        "statistics",
        "owned",
        "with_custom_attributes",
        "min_access_level",
    )
    _types = {"skip_groups": types.ArrayAttribute}


class GroupSubgroup(RESTObject):
    pass


class GroupSubgroupManager(SubgroupBaseManager[GroupSubgroup]):
    _path = "/groups/{group_id}/subgroups"
    _obj_cls = GroupSubgroup


class GroupDescendantGroup(RESTObject):
    pass


class GroupDescendantGroupManager(SubgroupBaseManager[GroupDescendantGroup]):
    """
    This manager inherits from GroupSubgroupManager as descendant groups
    share all attributes with subgroups, except the path and object class.
    """

    _path = "/groups/{group_id}/descendant_groups"
    _obj_cls = GroupDescendantGroup


class GroupLDAPGroupLink(RESTObject):
    _repr_attr = "provider"

    def _get_link_attrs(self) -> dict[str, str]:
        # https://docs.gitlab.com/ee/api/groups.html#add-ldap-group-link-with-cn-or-filter
        # https://docs.gitlab.com/ee/api/groups.html#delete-ldap-group-link-with-cn-or-filter
        # We can tell what attribute to use based on the data returned
        data = {"provider": self.provider}
        if self.cn:
            data["cn"] = self.cn
        else:
            data["filter"] = self.filter

        return data

    def delete(self, **kwargs: Any) -> None:
        """Delete the LDAP group link from the server.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabDeleteError: If the server cannot perform the request
        """
        if TYPE_CHECKING:
            assert isinstance(self.manager, DeleteMixin)
        self.manager.delete(
            self.encoded_id, query_data=self._get_link_attrs(), **kwargs
        )


class GroupLDAPGroupLinkManager(
    ListMixin[GroupLDAPGroupLink],
    CreateMixin[GroupLDAPGroupLink],
    DeleteMixin[GroupLDAPGroupLink],
):
    _path = "/groups/{group_id}/ldap_group_links"
    _obj_cls = GroupLDAPGroupLink
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = RequiredOptional(
        required=("provider", "group_access"), exclusive=("cn", "filter")
    )


class GroupSAMLGroupLink(ObjectDeleteMixin, RESTObject):
    _id_attr = "name"
    _repr_attr = "name"


class GroupSAMLGroupLinkManager(NoUpdateMixin[GroupSAMLGroupLink]):
    _path = "/groups/{group_id}/saml_group_links"
    _obj_cls = GroupSAMLGroupLink
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = RequiredOptional(required=("saml_group_name", "access_level"))



from __future__ import annotations

from gitlab.base import RESTObject
from gitlab.mixins import CreateMixin, ListMixin, RefreshMixin, RetrieveMixin
from gitlab.types import RequiredOptional

__all__ = [
    "BulkImport",
    "BulkImportManager",
    "BulkImportAllEntity",
    "BulkImportAllEntityManager",
    "BulkImportEntity",
    "BulkImportEntityManager",
]


class BulkImport(RefreshMixin, RESTObject):
    entities: BulkImportEntityManager


class BulkImportManager(CreateMixin[BulkImport], RetrieveMixin[BulkImport]):
    _path = "/bulk_imports"
    _obj_cls = BulkImport
    _create_attrs = RequiredOptional(required=("configuration", "entities"))
    _list_filters = ("sort", "status")


class BulkImportEntity(RefreshMixin, RESTObject):
    pass


class BulkImportEntityManager(RetrieveMixin[BulkImportEntity]):
    _path = "/bulk_imports/{bulk_import_id}/entities"
    _obj_cls = BulkImportEntity
    _from_parent_attrs = {"bulk_import_id": "id"}
    _list_filters = ("sort", "status")


class BulkImportAllEntity(RESTObject):
    pass


class BulkImportAllEntityManager(ListMixin[BulkImportAllEntity]):
    _path = "/bulk_imports/entities"
    _obj_cls = BulkImportAllEntity
    _list_filters = ("sort", "status")



from typing import Any, TYPE_CHECKING

from gitlab import cli
from gitlab import exceptions as exc
from gitlab.base import RESTObject
from gitlab.mixins import RetrieveMixin
from gitlab.utils import EncodedId

__all__ = ["Namespace", "NamespaceManager"]


class Namespace(RESTObject):
    pass


class NamespaceManager(RetrieveMixin[Namespace]):
    _path = "/namespaces"
    _obj_cls = Namespace
    _list_filters = ("search",)

    @cli.register_custom_action(
        cls_names="NamespaceManager", required=("namespace", "parent_id")
    )
    @exc.on_http_error(exc.GitlabGetError)
    def exists(self, namespace: str, **kwargs: Any) -> Namespace:
        """Get existence of a namespace by path.

        Args:
            namespace: The path to the namespace.
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the server failed to perform the request

        Returns:
            Data on namespace existence returned from the server.
        """
        path = f"{self.path}/{EncodedId(namespace)}/exists"
        server_data = self.gitlab.http_get(path, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(server_data, dict)
        return self._obj_cls(self, server_data)



"""Wrapper for the GitLab API."""

from __future__ import annotations

import os
import re
from typing import Any, BinaryIO, cast, TYPE_CHECKING
from urllib import parse

import requests

import gitlab
import gitlab.config
import gitlab.const
import gitlab.exceptions
from gitlab import _backends, utils

try:
    import gql
    import gql.transport.exceptions
    import httpx

    from ._backends.graphql import GitlabAsyncTransport, GitlabTransport

    _GQL_INSTALLED = True
except ImportError:  # pragma: no cover
    _GQL_INSTALLED = False


REDIRECT_MSG = (
    "python-gitlab detected a {status_code} ({reason!r}) redirection. You must update "
    "your GitLab URL to the correct URL to avoid issues. The redirection was from: "
    "{source!r} to {target!r}"
)


# https://docs.gitlab.com/ee/api/#offset-based-pagination
_PAGINATION_URL = (
    f"https://python-gitlab.readthedocs.io/en/v{gitlab.__version__}/"
    f"api-usage.html#pagination"
)


class Gitlab:
    """Represents a GitLab server connection.

    Args:
        url: The URL of the GitLab server (defaults to https://gitlab.com).
        private_token: The user private token
        oauth_token: An oauth token
        job_token: A CI job token
        ssl_verify: Whether SSL certificates should be validated. If
            the value is a string, it is the path to a CA file used for
            certificate validation.
        timeout: Timeout to use for requests to the GitLab server.
        http_username: Username for HTTP authentication
        http_password: Password for HTTP authentication
        api_version: Gitlab API version to use (support for 4 only)
        pagination: Can be set to 'keyset' to use keyset pagination
        order_by: Set order_by globally
        user_agent: A custom user agent to use for making HTTP requests.
        retry_transient_errors: Whether to retry after 500, 502, 503, 504
            or 52x responses. Defaults to False.
        keep_base_url: keep user-provided base URL for pagination if it
            differs from response headers

    Keyword Args:
        requests.Session session: HTTP Requests Session
        RequestsBackend backend: Backend that will be used to make http requests
    """

    def __init__(
        self,
        url: str | None = None,
        private_token: str | None = None,
        oauth_token: str | None = None,
        job_token: str | None = None,
        ssl_verify: bool | str = True,
        http_username: str | None = None,
        http_password: str | None = None,
        timeout: float | None = None,
        api_version: str = "4",
        per_page: int | None = None,
        pagination: str | None = None,
        order_by: str | None = None,
        user_agent: str = gitlab.const.USER_AGENT,
        retry_transient_errors: bool = False,
        keep_base_url: bool = False,
        **kwargs: Any,
    ) -> None:
        self._api_version = str(api_version)
        self._server_version: str | None = None
        self._server_revision: str | None = None
        self._base_url = utils.get_base_url(url)
        self._url = f"{self._base_url}/api/v{api_version}"
        #: Timeout to use for requests to gitlab server
        self.timeout = timeout
        self.retry_transient_errors = retry_transient_errors
        self.keep_base_url = keep_base_url
        #: Headers that will be used in request to GitLab
        self.headers = {"User-Agent": user_agent}

        #: Whether SSL certificates should be validated
        self.ssl_verify = ssl_verify

        self.private_token = private_token
        self.http_username = http_username
        self.http_password = http_password
        self.oauth_token = oauth_token
        self.job_token = job_token
        self._set_auth_info()

        #: Create a session object for requests
        _backend: type[_backends.DefaultBackend] = kwargs.pop(
            "backend", _backends.DefaultBackend
        )
        self._backend = _backend(**kwargs)
        self.session = self._backend.client

        self.per_page = per_page
        self.pagination = pagination
        self.order_by = order_by

        # We only support v4 API at this time
        if self._api_version not in ("4",):
            raise ModuleNotFoundError(f"gitlab.v{self._api_version}.objects")
        # NOTE: We must delay import of gitlab.v4.objects until now or
        # otherwise it will cause circular import errors
        from gitlab.v4 import objects

        self._objects = objects
        self.user: objects.CurrentUser | None = None

        self.broadcastmessages = objects.BroadcastMessageManager(self)
        """See :class:`~gitlab.v4.objects.BroadcastMessageManager`"""
        self.bulk_imports = objects.BulkImportManager(self)
        """See :class:`~gitlab.v4.objects.BulkImportManager`"""
        self.bulk_import_entities = objects.BulkImportAllEntityManager(self)
        """See :class:`~gitlab.v4.objects.BulkImportAllEntityManager`"""
        self.ci_lint = objects.CiLintManager(self)
        """See :class:`~gitlab.v4.objects.CiLintManager`"""
        self.deploykeys = objects.DeployKeyManager(self)
        """See :class:`~gitlab.v4.objects.DeployKeyManager`"""
        self.deploytokens = objects.DeployTokenManager(self)
        """See :class:`~gitlab.v4.objects.DeployTokenManager`"""
        self.geonodes = objects.GeoNodeManager(self)
        """See :class:`~gitlab.v4.objects.GeoNodeManager`"""
        self.gitlabciymls = objects.GitlabciymlManager(self)
        """See :class:`~gitlab.v4.objects.GitlabciymlManager`"""
        self.gitignores = objects.GitignoreManager(self)
        """See :class:`~gitlab.v4.objects.GitignoreManager`"""
        self.groups = objects.GroupManager(self)
        """See :class:`~gitlab.v4.objects.GroupManager`"""
        self.hooks = objects.HookManager(self)
        """See :class:`~gitlab.v4.objects.HookManager`"""
        self.issues = objects.IssueManager(self)
        """See :class:`~gitlab.v4.objects.IssueManager`"""
        self.issues_statistics = objects.IssuesStatisticsManager(self)
        """See :class:`~gitlab.v4.objects.IssuesStatisticsManager`"""
        self.keys = objects.KeyManager(self)
        """See :class:`~gitlab.v4.objects.KeyManager`"""
        self.ldapgroups = objects.LDAPGroupManager(self)
        """See :class:`~gitlab.v4.objects.LDAPGroupManager`"""
        self.licenses = objects.LicenseManager(self)
        """See :class:`~gitlab.v4.objects.LicenseManager`"""
        self.namespaces = objects.NamespaceManager(self)
        """See :class:`~gitlab.v4.objects.NamespaceManager`"""
        self.member_roles = objects.MemberRoleManager(self)
        """See :class:`~gitlab.v4.objects.MergeRequestManager`"""
        self.mergerequests = objects.MergeRequestManager(self)
        """See :class:`~gitlab.v4.objects.MergeRequestManager`"""
        self.notificationsettings = objects.NotificationSettingsManager(self)
        """See :class:`~gitlab.v4.objects.NotificationSettingsManager`"""
        self.projects = objects.ProjectManager(self)
        """See :class:`~gitlab.v4.objects.ProjectManager`"""
        self.registry_repositories = objects.RegistryRepositoryManager(self)
        """See :class:`~gitlab.v4.objects.RegistryRepositoryManager`"""
        self.runners = objects.RunnerManager(self)
        """See :class:`~gitlab.v4.objects.RunnerManager`"""
        self.runners_all = objects.RunnerAllManager(self)
        """See :class:`~gitlab.v4.objects.RunnerManager`"""
        self.settings = objects.ApplicationSettingsManager(self)
        """See :class:`~gitlab.v4.objects.ApplicationSettingsManager`"""
        self.appearance = objects.ApplicationAppearanceManager(self)
        """See :class:`~gitlab.v4.objects.ApplicationAppearanceManager`"""
        self.sidekiq = objects.SidekiqManager(self)
        """See :class:`~gitlab.v4.objects.SidekiqManager`"""
        self.snippets = objects.SnippetManager(self)
        """See :class:`~gitlab.v4.objects.SnippetManager`"""
        self.users = objects.UserManager(self)
        """See :class:`~gitlab.v4.objects.UserManager`"""
        self.todos = objects.TodoManager(self)
        """See :class:`~gitlab.v4.objects.TodoManager`"""
        self.dockerfiles = objects.DockerfileManager(self)
        """See :class:`~gitlab.v4.objects.DockerfileManager`"""
        self.events = objects.EventManager(self)
        """See :class:`~gitlab.v4.objects.EventManager`"""
        self.audit_events = objects.AuditEventManager(self)
        """See :class:`~gitlab.v4.objects.AuditEventManager`"""
        self.features = objects.FeatureManager(self)
        """See :class:`~gitlab.v4.objects.FeatureManager`"""
        self.pagesdomains = objects.PagesDomainManager(self)
        """See :class:`~gitlab.v4.objects.PagesDomainManager`"""
        self.user_activities = objects.UserActivitiesManager(self)
        """See :class:`~gitlab.v4.objects.UserActivitiesManager`"""
        self.applications = objects.ApplicationManager(self)
        """See :class:`~gitlab.v4.objects.ApplicationManager`"""
        self.variables = objects.VariableManager(self)
        """See :class:`~gitlab.v4.objects.VariableManager`"""
        self.personal_access_tokens = objects.PersonalAccessTokenManager(self)
        """See :class:`~gitlab.v4.objects.PersonalAccessTokenManager`"""
        self.topics = objects.TopicManager(self)
        """See :class:`~gitlab.v4.objects.TopicManager`"""
        self.service_accounts = objects.ServiceAccountManager(self)
        """See :class:`~gitlab.v4.objects.ServiceAccountManager`"""
        self.statistics = objects.ApplicationStatisticsManager(self)
        """See :class:`~gitlab.v4.objects.ApplicationStatisticsManager`"""

    def __enter__(self) -> Gitlab:
        return self

    def __exit__(self, *args: Any) -> None:
        self.session.close()

    def __getstate__(self) -> dict[str, Any]:
        state = self.__dict__.copy()
        state.pop("_objects")
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        self.__dict__.update(state)
        # We only support v4 API at this time
        if self._api_version not in ("4",):
            raise ModuleNotFoundError(
                f"gitlab.v{self._api_version}.objects"
            )  # pragma: no cover, dead code currently
        # NOTE: We must delay import of gitlab.v4.objects until now or
        # otherwise it will cause circular import errors
        from gitlab.v4 import objects

        self._objects = objects

    @property
    def url(self) -> str:
        """The user-provided server URL."""
        return self._base_url

    @property
    def api_url(self) -> str:
        """The computed API base URL."""
        return self._url

    @property
    def api_version(self) -> str:
        """The API version used (4 only)."""
        return self._api_version

    @classmethod
    def from_config(
        cls,
        gitlab_id: str | None = None,
        config_files: list[str] | None = None,
        **kwargs: Any,
    ) -> Gitlab:
        """Create a Gitlab connection from configuration files.

        Args:
            gitlab_id: ID of the configuration section.
            config_files list[str]: List of paths to configuration files.

        kwargs:
            session requests.Session: Custom requests Session

        Returns:
            A Gitlab connection.

        Raises:
            gitlab.config.GitlabDataError: If the configuration is not correct.
        """
        config = gitlab.config.GitlabConfigParser(
            gitlab_id=gitlab_id, config_files=config_files
        )
        return cls(
            config.url,
            private_token=config.private_token,
            oauth_token=config.oauth_token,
            job_token=config.job_token,
            ssl_verify=config.ssl_verify,
            timeout=config.timeout,
            http_username=config.http_username,
            http_password=config.http_password,
            api_version=config.api_version,
            per_page=config.per_page,
            pagination=config.pagination,
            order_by=config.order_by,
            user_agent=config.user_agent,
            retry_transient_errors=config.retry_transient_errors,
            keep_base_url=config.keep_base_url,
            **kwargs,
        )

    @classmethod
    def merge_config(
        cls,
        options: dict[str, Any],
        gitlab_id: str | None = None,
        config_files: list[str] | None = None,
    ) -> Gitlab:
        """Create a Gitlab connection by merging configuration with
        the following precedence:

        1. Explicitly provided CLI arguments,
        2. Environment variables,
        3. Configuration files:
            a. explicitly defined config files:
                i. via the `--config-file` CLI argument,
                ii. via the `PYTHON_GITLAB_CFG` environment variable,
            b. user-specific config file,
            c. system-level config file,
        4. Environment variables always present in CI (CI_SERVER_URL, CI_JOB_TOKEN).

        Args:
            options: A dictionary of explicitly provided key-value options.
            gitlab_id: ID of the configuration section.
            config_files: List of paths to configuration files.
        Returns:
            (gitlab.Gitlab): A Gitlab connection.

        Raises:
            gitlab.config.GitlabDataError: If the configuration is not correct.
        """
        config = gitlab.config.GitlabConfigParser(
            gitlab_id=gitlab_id, config_files=config_files
        )
        url = (
            options.get("server_url")
            or config.url
            or os.getenv("CI_SERVER_URL")
            or gitlab.const.DEFAULT_URL
        )
        private_token, oauth_token, job_token = cls._merge_auth(options, config)

        return cls(
            url=url,
            private_token=private_token,
            oauth_token=oauth_token,
            job_token=job_token,
            ssl_verify=options.get("ssl_verify") or config.ssl_verify,
            timeout=options.get("timeout") or config.timeout,
            api_version=options.get("api_version") or config.api_version,
            per_page=options.get("per_page") or config.per_page,
            pagination=options.get("pagination") or config.pagination,
            order_by=options.get("order_by") or config.order_by,
            user_agent=options.get("user_agent") or config.user_agent,
        )

    @staticmethod
    def _merge_auth(
        options: dict[str, Any], config: gitlab.config.GitlabConfigParser
    ) -> tuple[str | None, str | None, str | None]:
        """
        Return a tuple where at most one of 3 token types ever has a value.
        Since multiple types of tokens may be present in the environment,
        options, or config files, this precedence ensures we don't
        inadvertently cause errors when initializing the client.

        This is especially relevant when executed in CI where user and
        CI-provided values are both available.
        """
        private_token = options.get("private_token") or config.private_token
        oauth_token = options.get("oauth_token") or config.oauth_token
        job_token = (
            options.get("job_token") or config.job_token or os.getenv("CI_JOB_TOKEN")
        )

        if private_token:
            return (private_token, None, None)
        if oauth_token:
            return (None, oauth_token, None)
        if job_token:
            return (None, None, job_token)

        return (None, None, None)

    def auth(self) -> None:
        """Performs an authentication using private token. Warns the user if a
        potentially misconfigured URL is detected on the client or server side.

        The `user` attribute will hold a `gitlab.objects.CurrentUser` object on
        success.
        """
        self.user = self._objects.CurrentUserManager(self).get()

        if hasattr(self.user, "web_url") and hasattr(self.user, "username"):
            self._check_url(self.user.web_url, path=self.user.username)

    def version(self) -> tuple[str, str]:
        """Returns the version and revision of the gitlab server.

        Note that self.version and self.revision will be set on the gitlab
        object.

        Returns:
            The server version and server revision.
                ('unknown', 'unknown') if the server doesn't perform as expected.
        """
        if self._server_version is None:
            try:
                data = self.http_get("/version")
                if isinstance(data, dict):
                    self._server_version = data["version"]
                    self._server_revision = data["revision"]
                else:
                    self._server_version = "unknown"
                    self._server_revision = "unknown"
            except Exception:
                self._server_version = "unknown"
                self._server_revision = "unknown"

        return cast(str, self._server_version), cast(str, self._server_revision)

    @gitlab.exceptions.on_http_error(gitlab.exceptions.GitlabMarkdownError)
    def markdown(
        self, text: str, gfm: bool = False, project: str | None = None, **kwargs: Any
    ) -> str:
        """Render an arbitrary Markdown document.

        Args:
            text: The markdown text to render
            gfm: Render text using GitLab Flavored Markdown. Default is False
            project: Full path of a project used a context when `gfm` is True
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabMarkdownError: If the server cannot perform the request

        Returns:
            The HTML rendering of the markdown text.
        """
        post_data = {"text": text, "gfm": gfm}
        if project is not None:
            post_data["project"] = project
        data = self.http_post("/markdown", post_data=post_data, **kwargs)
        if TYPE_CHECKING:
            assert not isinstance(data, requests.Response)
            assert isinstance(data["html"], str)
        return data["html"]

    @gitlab.exceptions.on_http_error(gitlab.exceptions.GitlabLicenseError)
    def get_license(self, **kwargs: Any) -> dict[str, str | dict[str, str]]:
        """Retrieve information about the current license.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the server cannot perform the request

        Returns:
            The current license information
        """
        result = self.http_get("/license", **kwargs)
        if isinstance(result, dict):
            return result
        return {}

    @gitlab.exceptions.on_http_error(gitlab.exceptions.GitlabLicenseError)
    def set_license(self, license: str, **kwargs: Any) -> dict[str, Any]:
        """Add a new license.

        Args:
            license: The license string
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabPostError: If the server cannot perform the request

        Returns:
            The new license information
        """
        data = {"license": license}
        result = self.http_post("/license", post_data=data, **kwargs)
        if TYPE_CHECKING:
            assert not isinstance(result, requests.Response)
        return result

    def _set_auth_info(self) -> None:
        tokens = [
            token
            for token in [self.private_token, self.oauth_token, self.job_token]
            if token
        ]
        if len(tokens) > 1:
            raise ValueError(
                "Only one of private_token, oauth_token or job_token should "
                "be defined"
            )
        if (self.http_username and not self.http_password) or (
            not self.http_username and self.http_password
        ):
            raise ValueError("Both http_username and http_password should be defined")
        if tokens and self.http_username:
            raise ValueError(
                "Only one of token authentications or http "
                "authentication should be defined"
            )

        self._auth: requests.auth.AuthBase | None = None
        if self.private_token:
            self._auth = _backends.PrivateTokenAuth(self.private_token)

        if self.oauth_token:
            self._auth = _backends.OAuthTokenAuth(self.oauth_token)

        if self.job_token:
            self._auth = _backends.JobTokenAuth(self.job_token)

        if self.http_username and self.http_password:
            self._auth = requests.auth.HTTPBasicAuth(
                self.http_username, self.http_password
            )

    def enable_debug(self, mask_credentials: bool = True) -> None:
        import logging
        from http import client

        client.HTTPConnection.debuglevel = 1
        logging.basicConfig()
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)

        httpclient_log = logging.getLogger("http.client")
        httpclient_log.propagate = True
        httpclient_log.setLevel(logging.DEBUG)

        requests_log = logging.getLogger("requests.packages.urllib3")
        requests_log.setLevel(logging.DEBUG)
        requests_log.propagate = True

        # shadow http.client prints to log()
        # https://stackoverflow.com/a/16337639
        def print_as_log(*args: Any) -> None:
            httpclient_log.log(logging.DEBUG, " ".join(args))

        setattr(client, "print", print_as_log)

        if not mask_credentials:
            return

        token = self.private_token or self.oauth_token or self.job_token
        handler = logging.StreamHandler()
        handler.setFormatter(utils.MaskingFormatter(masked=token))
        logger.handlers.clear()
        logger.addHandler(handler)

    def _get_session_opts(self) -> dict[str, Any]:
        return {
            "headers": self.headers.copy(),
            "auth": self._auth,
            "timeout": self.timeout,
            "verify": self.ssl_verify,
        }

    def _build_url(self, path: str) -> str:
        """Returns the full url from path.

        If path is already a url, return it unchanged. If it's a path, append
        it to the stored url.

        Returns:
            The full URL
        """
        if path.startswith("http://") or path.startswith("https://"):
            return path
        return f"{self._url}{path}"

    def _check_url(self, url: str | None, *, path: str = "api") -> str | None:
        """
        Checks if ``url`` starts with a different base URL from the user-provided base
        URL and warns the user before returning it. If ``keep_base_url`` is set to
        ``True``, instead returns the URL massaged to match the user-provided base URL.
        """
        if not url or url.startswith(self.url):
            return url

        match = re.match(rf"(^.*?)/{path}", url)
        if not match:
            return url

        base_url = match.group(1)
        if self.keep_base_url:
            return url.replace(base_url, f"{self._base_url}")

        utils.warn(
            message=(
                f"The base URL in the server response differs from the user-provided "
                f"base URL ({self.url} -> {base_url}).\nThis is usually caused by a "
                f"misconfigured base URL on your side or a misconfigured external_url "
                f"on the server side, and can lead to broken pagination and unexpected "
                f"behavior. If this is intentional, use `keep_base_url=True` when "
                f"initializing the Gitlab instance to keep the user-provided base URL."
            ),
            category=UserWarning,
        )
        return url

    @staticmethod
    def _check_redirects(result: requests.Response) -> None:
        # Check the requests history to detect 301/302 redirections.
        # If the initial verb is POST or PUT, the redirected request will use a
        # GET request, leading to unwanted behaviour.
        # If we detect a redirection with a POST or a PUT request, we
        # raise an exception with a useful error message.
        if not result.history:
            return

        for item in result.history:
            if item.status_code not in (301, 302):
                continue
            # GET and HEAD methods can be redirected without issue
            if item.request.method in ("GET", "HEAD"):
                continue
            target = item.headers.get("location")
            raise gitlab.exceptions.RedirectError(
                REDIRECT_MSG.format(
                    status_code=item.status_code,
                    reason=item.reason,
                    source=item.url,
                    target=target,
                )
            )

    def http_request(
        self,
        verb: str,
        path: str,
        query_data: dict[str, Any] | None = None,
        post_data: dict[str, Any] | bytes | BinaryIO | None = None,
        raw: bool = False,
        streamed: bool = False,
        files: dict[str, Any] | None = None,
        timeout: float | None = None,
        obey_rate_limit: bool = True,
        retry_transient_errors: bool | None = None,
        max_retries: int = 10,
        extra_headers: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> requests.Response:
        """Make an HTTP request to the Gitlab server.

        Args:
            verb: The HTTP method to call ('get', 'post', 'put', 'delete')
            path: Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projecs')
            query_data: Data to send as query parameters
            post_data: Data to send in the body (will be converted to
                              json by default)
            raw: If True, do not convert post_data to json
            streamed: Whether the data should be streamed
            files: The files to send to the server
            timeout: The timeout, in seconds, for the request
            obey_rate_limit: Whether to obey 429 Too Many Request
                                    responses. Defaults to True.
            retry_transient_errors: Whether to retry after 500, 502, 503, 504
                or 52x responses. Defaults to False.
            max_retries: Max retries after 429 or transient errors,
                               set to -1 to retry forever. Defaults to 10.
            extra_headers: Add and override HTTP headers for the request.
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            A requests result object.

        Raises:
            GitlabHttpError: When the return code is not 2xx
        """
        query_data = query_data or {}
        raw_url = self._build_url(path)

        # parse user-provided URL params to ensure we don't add our own duplicates
        parsed = parse.urlparse(raw_url)
        params = parse.parse_qs(parsed.query)
        utils.copy_dict(src=query_data, dest=params)

        url = parse.urlunparse(parsed._replace(query=""))

        # Deal with kwargs: by default a user uses kwargs to send data to the
        # gitlab server, but this generates problems (python keyword conflicts
        # and python-gitlab/gitlab conflicts).
        # So we provide a `query_parameters` key: if it's there we use its dict
        # value as arguments for the gitlab server, and ignore the other
        # arguments, except pagination ones (per_page and page)
        if "query_parameters" in kwargs:
            utils.copy_dict(src=kwargs["query_parameters"], dest=params)
            for arg in ("per_page", "page"):
                if arg in kwargs:
                    params[arg] = kwargs[arg]
        else:
            utils.copy_dict(src=kwargs, dest=params)

        opts = self._get_session_opts()

        verify = opts.pop("verify")
        opts_timeout = opts.pop("timeout")
        # If timeout was passed into kwargs, allow it to override the default
        if timeout is None:
            timeout = opts_timeout
        if retry_transient_errors is None:
            retry_transient_errors = self.retry_transient_errors

        # We need to deal with json vs. data when uploading files
        send_data = self._backend.prepare_send_data(files, post_data, raw)
        opts["headers"]["Content-type"] = send_data.content_type

        if extra_headers is not None:
            opts["headers"].update(extra_headers)

        retry = utils.Retry(
            max_retries=max_retries,
            obey_rate_limit=obey_rate_limit,
            retry_transient_errors=retry_transient_errors,
        )

        while True:
            try:
                result = self._backend.http_request(
                    method=verb,
                    url=url,
                    json=send_data.json,
                    data=send_data.data,
                    params=params,
                    timeout=timeout,
                    verify=verify,
                    stream=streamed,
                    **opts,
                )
            except (requests.ConnectionError, requests.exceptions.ChunkedEncodingError):
                if retry.handle_retry():
                    continue
                raise

            self._check_redirects(result.response)

            if 200 <= result.status_code < 300:
                return result.response

            if retry.handle_retry_on_status(
                result.status_code, result.headers, result.reason
            ):
                continue

            error_message = result.content
            try:
                error_json = result.json()
                for k in ("message", "error"):
                    if k in error_json:
                        error_message = error_json[k]
            except (KeyError, ValueError, TypeError):
                pass

            if result.status_code == 401:
                raise gitlab.exceptions.GitlabAuthenticationError(
                    response_code=result.status_code,
                    error_message=error_message,
                    response_body=result.content,
                )

            raise gitlab.exceptions.GitlabHttpError(
                response_code=result.status_code,
                error_message=error_message,
                response_body=result.content,
            )

    def http_get(
        self,
        path: str,
        query_data: dict[str, Any] | None = None,
        streamed: bool = False,
        raw: bool = False,
        **kwargs: Any,
    ) -> dict[str, Any] | requests.Response:
        """Make a GET request to the Gitlab server.

        Args:
            path: Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projecs')
            query_data: Data to send as query parameters
            streamed: Whether the data should be streamed
            raw: If True do not try to parse the output as json
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            A requests result object is streamed is True or the content type is
            not json.
            The parsed json data otherwise.

        Raises:
            GitlabHttpError: When the return code is not 2xx
            GitlabParsingError: If the json data could not be parsed
        """
        query_data = query_data or {}
        result = self.http_request(
            "get", path, query_data=query_data, streamed=streamed, **kwargs
        )
        content_type = utils.get_content_type(result.headers.get("Content-Type"))

        if content_type == "application/json" and not streamed and not raw:
            try:
                json_result = result.json()
                if TYPE_CHECKING:
                    assert isinstance(json_result, dict)
                return json_result
            except Exception as e:
                raise gitlab.exceptions.GitlabParsingError(
                    error_message="Failed to parse the server message"
                ) from e
        else:
            return result

    def http_head(
        self, path: str, query_data: dict[str, Any] | None = None, **kwargs: Any
    ) -> requests.structures.CaseInsensitiveDict[Any]:
        """Make a HEAD request to the Gitlab server.

        Args:
            path: Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projecs')
            query_data: Data to send as query parameters
            **kwargs: Extra options to send to the server (e.g. sudo, page,
                      per_page)
        Returns:
            A requests.header object
        Raises:
            GitlabHttpError: When the return code is not 2xx
        """

        query_data = query_data or {}
        result = self.http_request("head", path, query_data=query_data, **kwargs)
        return result.headers

    def http_list(
        self,
        path: str,
        query_data: dict[str, Any] | None = None,
        *,
        iterator: bool | None = None,
        message_details: utils.WarnMessageData | None = None,
        **kwargs: Any,
    ) -> GitlabList | list[dict[str, Any]]:
        """Make a GET request to the Gitlab server for list-oriented queries.

        Args:
            path: Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projects')
            query_data: Data to send as query parameters
            iterator: Indicate if should return a generator (True)
            **kwargs: Extra options to send to the server (e.g. sudo, page,
                      per_page)

        Returns:
            A list of the objects returned by the server. If `iterator` is
            True and no pagination-related arguments (`page`, `per_page`,
            `get_all`) are defined then a GitlabList object (generator) is returned
            instead. This object will make API calls when needed to fetch the
            next items from the server.

        Raises:
            GitlabHttpError: When the return code is not 2xx
            GitlabParsingError: If the json data could not be parsed
        """
        query_data = query_data or {}

        # Provide a `get_all`` param to avoid clashes with `all` API attributes.
        get_all = kwargs.pop("get_all", None)

        if get_all is None:
            # For now, keep `all` without deprecation.
            get_all = kwargs.pop("all", None)

        url = self._build_url(path)

        page = kwargs.get("page")

        if iterator:
            if page is not None:
                utils.warn(
                    message=(
                        f"`{iterator=}` and `{page=}` were both specified. "
                        f"`{page=}` will be ignored."
                    ),
                    category=UserWarning,
                )

            # Generator requested
            return GitlabList(self, url, query_data, **kwargs)

        if get_all is True:
            return list(GitlabList(self, url, query_data, **kwargs))

        # pagination requested, we return a list
        gl_list = GitlabList(self, url, query_data, get_next=False, **kwargs)
        items = list(gl_list)

        def should_emit_warning() -> bool:
            # No warning is emitted if any of the following conditions apply:
            # * `get_all=False` was set in the `list()` call.
            # * `page` was set in the `list()` call.
            # * GitLab did not return the `x-per-page` header.
            # * Number of items received is less than per-page value.
            # * Number of items received is >= total available.
            if get_all is False:
                return False
            if page is not None:
                return False
            if gl_list.per_page is None:
                return False
            if len(items) < gl_list.per_page:
                return False
            if gl_list.total is not None and len(items) >= gl_list.total:
                return False
            return True

        if not should_emit_warning():
            return items

        # Warn the user that they are only going to retrieve `per_page`
        # maximum items. This is a common cause of issues filed.
        total_items = "many" if gl_list.total is None else gl_list.total
        if message_details is not None:
            message = message_details.message.format_map(
                {
                    "len_items": len(items),
                    "per_page": gl_list.per_page,
                    "total_items": total_items,
                }
            )
            show_caller = message_details.show_caller
        else:
            message = (
                f"Calling a `list()` method without specifying `get_all=True` or "
                f"`iterator=True` will return a maximum of {gl_list.per_page} items. "
                f"Your query returned {len(items)} of {total_items} items. See "
                f"{_PAGINATION_URL} for more details. If this was done intentionally, "
                f"then this warning can be supressed by adding the argument "
                f"`get_all=False` to the `list()` call."
            )
            show_caller = True
        utils.warn(message=message, category=UserWarning, show_caller=show_caller)
        return items

    def http_post(
        self,
        path: str,
        query_data: dict[str, Any] | None = None,
        post_data: dict[str, Any] | None = None,
        raw: bool = False,
        files: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> dict[str, Any] | requests.Response:
        """Make a POST request to the Gitlab server.

        Args:
            path: Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projecs')
            query_data: Data to send as query parameters
            post_data: Data to send in the body (will be converted to
                              json by default)
            raw: If True, do not convert post_data to json
            files: The files to send to the server
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            The parsed json returned by the server if json is return, else the
            raw content

        Raises:
            GitlabHttpError: When the return code is not 2xx
            GitlabParsingError: If the json data could not be parsed
        """
        query_data = query_data or {}
        post_data = post_data or {}

        result = self.http_request(
            "post",
            path,
            query_data=query_data,
            post_data=post_data,
            files=files,
            raw=raw,
            **kwargs,
        )
        content_type = utils.get_content_type(result.headers.get("Content-Type"))

        try:
            if content_type == "application/json":
                json_result = result.json()
                if TYPE_CHECKING:
                    assert isinstance(json_result, dict)
                return json_result
        except Exception as e:
            raise gitlab.exceptions.GitlabParsingError(
                error_message="Failed to parse the server message"
            ) from e
        return result

    def http_put(
        self,
        path: str,
        query_data: dict[str, Any] | None = None,
        post_data: dict[str, Any] | bytes | BinaryIO | None = None,
        raw: bool = False,
        files: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> dict[str, Any] | requests.Response:
        """Make a PUT request to the Gitlab server.

        Args:
            path: Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projecs')
            query_data: Data to send as query parameters
            post_data: Data to send in the body (will be converted to
                              json by default)
            raw: If True, do not convert post_data to json
            files: The files to send to the server
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            The parsed json returned by the server.

        Raises:
            GitlabHttpError: When the return code is not 2xx
            GitlabParsingError: If the json data could not be parsed
        """
        query_data = query_data or {}
        post_data = post_data or {}

        result = self.http_request(
            "put",
            path,
            query_data=query_data,
            post_data=post_data,
            files=files,
            raw=raw,
            **kwargs,
        )
        if result.status_code in gitlab.const.NO_JSON_RESPONSE_CODES:
            return result
        try:
            json_result = result.json()
            if TYPE_CHECKING:
                assert isinstance(json_result, dict)
            return json_result
        except Exception as e:
            raise gitlab.exceptions.GitlabParsingError(
                error_message="Failed to parse the server message"
            ) from e

    def http_patch(
        self,
        path: str,
        *,
        query_data: dict[str, Any] | None = None,
        post_data: dict[str, Any] | bytes | None = None,
        raw: bool = False,
        **kwargs: Any,
    ) -> dict[str, Any] | requests.Response:
        """Make a PATCH request to the Gitlab server.

        Args:
            path: Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projecs')
            query_data: Data to send as query parameters
            post_data: Data to send in the body (will be converted to
                              json by default)
            raw: If True, do not convert post_data to json
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            The parsed json returned by the server.

        Raises:
            GitlabHttpError: When the return code is not 2xx
            GitlabParsingError: If the json data could not be parsed
        """
        query_data = query_data or {}
        post_data = post_data or {}

        result = self.http_request(
            "patch", path, query_data=query_data, post_data=post_data, raw=raw, **kwargs
        )
        if result.status_code in gitlab.const.NO_JSON_RESPONSE_CODES:
            return result
        try:
            json_result = result.json()
            if TYPE_CHECKING:
                assert isinstance(json_result, dict)
            return json_result
        except Exception as e:
            raise gitlab.exceptions.GitlabParsingError(
                error_message="Failed to parse the server message"
            ) from e

    def http_delete(self, path: str, **kwargs: Any) -> requests.Response:
        """Make a DELETE request to the Gitlab server.

        Args:
            path: Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projecs')
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            The requests object.

        Raises:
            GitlabHttpError: When the return code is not 2xx
        """
        return self.http_request("delete", path, **kwargs)

    @gitlab.exceptions.on_http_error(gitlab.exceptions.GitlabSearchError)
    def search(
        self, scope: str, search: str, **kwargs: Any
    ) -> GitlabList | list[dict[str, Any]]:
        """Search GitLab resources matching the provided string.'

        Args:
            scope: Scope of the search
            search: Search string
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabSearchError: If the server failed to perform the request

        Returns:
            A list of dicts describing the resources found.
        """
        data = {"scope": scope, "search": search}
        return self.http_list("/search", query_data=data, **kwargs)


class GitlabList:
    """Generator representing a list of remote objects.

    The object handles the links returned by a query to the API, and will call
    the API again when needed.
    """

    def __init__(
        self,
        gl: Gitlab,
        url: str,
        query_data: dict[str, Any],
        get_next: bool = True,
        **kwargs: Any,
    ) -> None:
        self._gl = gl

        # Preserve kwargs for subsequent queries
        self._kwargs = kwargs.copy()

        self._query(url, query_data, **self._kwargs)
        self._get_next = get_next

        # Remove query_parameters from kwargs, which are saved via the `next` URL
        self._kwargs.pop("query_parameters", None)

    def _query(
        self, url: str, query_data: dict[str, Any] | None = None, **kwargs: Any
    ) -> None:
        query_data = query_data or {}
        result = self._gl.http_request("get", url, query_data=query_data, **kwargs)
        try:
            next_url = result.links["next"]["url"]
        except KeyError:
            next_url = None

        self._next_url = self._gl._check_url(next_url)
        self._current_page: str | None = result.headers.get("X-Page")
        self._prev_page: str | None = result.headers.get("X-Prev-Page")
        self._next_page: str | None = result.headers.get("X-Next-Page")
        self._per_page: str | None = result.headers.get("X-Per-Page")
        self._total_pages: str | None = result.headers.get("X-Total-Pages")
        self._total: str | None = result.headers.get("X-Total")

        try:
            self._data: list[dict[str, Any]] = result.json()
        except Exception as e:
            raise gitlab.exceptions.GitlabParsingError(
                error_message="Failed to parse the server message"
            ) from e

        self._current = 0

    @property
    def current_page(self) -> int:
        """The current page number."""
        if TYPE_CHECKING:
            assert self._current_page is not None
        return int(self._current_page)

    @property
    def prev_page(self) -> int | None:
        """The previous page number.

        If None, the current page is the first.
        """
        return int(self._prev_page) if self._prev_page else None

    @property
    def next_page(self) -> int | None:
        """The next page number.

        If None, the current page is the last.
        """
        return int(self._next_page) if self._next_page else None

    @property
    def per_page(self) -> int | None:
        """The number of items per page."""
        return int(self._per_page) if self._per_page is not None else None

    # NOTE(jlvillal): When a query returns more than 10,000 items, GitLab doesn't return
    # the headers 'x-total-pages' and 'x-total'. In those cases we return None.
    # https://docs.gitlab.com/ee/user/gitlab_com/index.html#pagination-response-headers
    @property
    def total_pages(self) -> int | None:
        """The total number of pages."""
        if self._total_pages is not None:
            return int(self._total_pages)
        return None

    @property
    def total(self) -> int | None:
        """The total number of items."""
        if self._total is not None:
            return int(self._total)
        return None

    def __iter__(self) -> GitlabList:
        return self

    def __len__(self) -> int:
        if self._total is None:
            return 0
        return int(self._total)

    def __next__(self) -> dict[str, Any]:
        return self.next()

    def next(self) -> dict[str, Any]:
        try:
            item = self._data[self._current]
            self._current += 1
            return item
        except IndexError:
            pass

        if self._next_url and self._get_next is True:
            self._query(self._next_url, **self._kwargs)
            return self.next()

        raise StopIteration


class _BaseGraphQL:
    def __init__(
        self,
        url: str | None = None,
        *,
        token: str | None = None,
        ssl_verify: bool | str = True,
        timeout: float | None = None,
        user_agent: str = gitlab.const.USER_AGENT,
        fetch_schema_from_transport: bool = False,
        max_retries: int = 10,
        obey_rate_limit: bool = True,
        retry_transient_errors: bool = False,
    ) -> None:
        if not _GQL_INSTALLED:
            raise ImportError(
                "The GraphQL client could not be initialized because "
                "the gql dependencies are not installed. "
                "Install them with 'pip install python-gitlab[graphql]'"
            )
        self._base_url = utils.get_base_url(url)
        self._timeout = timeout
        self._token = token
        self._url = f"{self._base_url}/api/graphql"
        self._user_agent = user_agent
        self._ssl_verify = ssl_verify
        self._max_retries = max_retries
        self._obey_rate_limit = obey_rate_limit
        self._retry_transient_errors = retry_transient_errors
        self._client_opts = self._get_client_opts()
        self._fetch_schema_from_transport = fetch_schema_from_transport

    def _get_client_opts(self) -> dict[str, Any]:
        headers = {"User-Agent": self._user_agent}

        if self._token:
            headers["Authorization"] = f"Bearer {self._token}"

        return {
            "headers": headers,
            "timeout": self._timeout,
            "verify": self._ssl_verify,
        }


class GraphQL(_BaseGraphQL):
    def __init__(
        self,
        url: str | None = None,
        *,
        token: str | None = None,
        ssl_verify: bool | str = True,
        client: httpx.Client | None = None,
        timeout: float | None = None,
        user_agent: str = gitlab.const.USER_AGENT,
        fetch_schema_from_transport: bool = False,
        max_retries: int = 10,
        obey_rate_limit: bool = True,
        retry_transient_errors: bool = False,
    ) -> None:
        super().__init__(
            url=url,
            token=token,
            ssl_verify=ssl_verify,
            timeout=timeout,
            user_agent=user_agent,
            fetch_schema_from_transport=fetch_schema_from_transport,
            max_retries=max_retries,
            obey_rate_limit=obey_rate_limit,
            retry_transient_errors=retry_transient_errors,
        )

        self._http_client = client or httpx.Client(**self._client_opts)
        self._transport = GitlabTransport(self._url, client=self._http_client)
        self._client = gql.Client(
            transport=self._transport,
            fetch_schema_from_transport=fetch_schema_from_transport,
        )
        self._gql = gql.gql

    def __enter__(self) -> GraphQL:
        return self

    def __exit__(self, *args: Any) -> None:
        self._http_client.close()

    def execute(self, request: str, *args: Any, **kwargs: Any) -> Any:
        parsed_document = self._gql(request)
        retry = utils.Retry(
            max_retries=self._max_retries,
            obey_rate_limit=self._obey_rate_limit,
            retry_transient_errors=self._retry_transient_errors,
        )

        while True:
            try:
                result = self._client.execute(parsed_document, *args, **kwargs)
            except gql.transport.exceptions.TransportServerError as e:
                if retry.handle_retry_on_status(
                    status_code=e.code, headers=self._transport.response_headers
                ):
                    continue

                if e.code == 401:
                    raise gitlab.exceptions.GitlabAuthenticationError(
                        response_code=e.code, error_message=str(e)
                    )

                raise gitlab.exceptions.GitlabHttpError(
                    response_code=e.code, error_message=str(e)
                )

            return result


class AsyncGraphQL(_BaseGraphQL):
    def __init__(
        self,
        url: str | None = None,
        *,
        token: str | None = None,
        ssl_verify: bool | str = True,
        client: httpx.AsyncClient | None = None,
        timeout: float | None = None,
        user_agent: str = gitlab.const.USER_AGENT,
        fetch_schema_from_transport: bool = False,
        max_retries: int = 10,
        obey_rate_limit: bool = True,
        retry_transient_errors: bool = False,
    ) -> None:
        super().__init__(
            url=url,
            token=token,
            ssl_verify=ssl_verify,
            timeout=timeout,
            user_agent=user_agent,
            fetch_schema_from_transport=fetch_schema_from_transport,
            max_retries=max_retries,
            obey_rate_limit=obey_rate_limit,
            retry_transient_errors=retry_transient_errors,
        )

        self._http_client = client or httpx.AsyncClient(**self._client_opts)
        self._transport = GitlabAsyncTransport(self._url, client=self._http_client)
        self._client = gql.Client(
            transport=self._transport,
            fetch_schema_from_transport=fetch_schema_from_transport,
        )
        self._gql = gql.gql

    async def __aenter__(self) -> AsyncGraphQL:
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self._http_client.aclose()

    async def execute(self, request: str, *args: Any, **kwargs: Any) -> Any:
        parsed_document = self._gql(request)
        retry = utils.Retry(
            max_retries=self._max_retries,
            obey_rate_limit=self._obey_rate_limit,
            retry_transient_errors=self._retry_transient_errors,
        )

        while True:
            try:
                result = await self._client.execute_async(
                    parsed_document, *args, **kwargs
                )
            except gql.transport.exceptions.TransportServerError as e:
                if retry.handle_retry_on_status(
                    status_code=e.code, headers=self._transport.response_headers
                ):
                    continue

                if e.code == 401:
                    raise gitlab.exceptions.GitlabAuthenticationError(
                        response_code=e.code, error_message=str(e)
                    )

                raise gitlab.exceptions.GitlabHttpError(
                    response_code=e.code, error_message=str(e)
                )

            return result



from __future__ import annotations

import argparse
import json
import operator
import sys
from typing import Any, TYPE_CHECKING

import gitlab
import gitlab.base
import gitlab.v4.objects
from gitlab import cli
from gitlab.exceptions import GitlabCiLintError


class GitlabCLI:
    def __init__(
        self,
        gl: gitlab.Gitlab,
        gitlab_resource: str,
        resource_action: str,
        args: dict[str, str],
    ) -> None:
        self.cls: type[gitlab.base.RESTObject] = cli.gitlab_resource_to_cls(
            gitlab_resource, namespace=gitlab.v4.objects
        )
        self.cls_name = self.cls.__name__
        self.gitlab_resource = gitlab_resource.replace("-", "_")
        self.resource_action = resource_action.lower()
        self.gl = gl
        self.args = args
        self.parent_args: dict[str, Any] = {}
        self.mgr_cls: Any = getattr(gitlab.v4.objects, f"{self.cls.__name__}Manager")
        # We could do something smart, like splitting the manager name to find
        # parents, build the chain of managers to get to the final object.
        # Instead we do something ugly and efficient: interpolate variables in
        # the class _path attribute, and replace the value with the result.

        self._process_from_parent_attrs()

        self.mgr_cls._path = self.mgr_cls._path.format(**self.parent_args)
        self.mgr: Any = self.mgr_cls(gl)
        self.mgr._from_parent_attrs = self.parent_args
        if self.mgr_cls._types:
            for attr_name, type_cls in self.mgr_cls._types.items():
                if attr_name in self.args.keys():
                    obj = type_cls()
                    obj.set_from_cli(self.args[attr_name])
                    self.args[attr_name] = obj.get()

    def _process_from_parent_attrs(self) -> None:
        """Items in the path need to be url-encoded. There is a 1:1 mapping from
        mgr_cls._from_parent_attrs <--> mgr_cls._path. Those values must be url-encoded
        as they may contain a slash '/'."""
        for key in self.mgr_cls._from_parent_attrs:
            if key not in self.args:
                continue

            self.parent_args[key] = gitlab.utils.EncodedId(self.args[key])
            # If we don't delete it then it will be added to the URL as a query-string
            del self.args[key]

    def run(self) -> Any:
        # Check for a method that matches gitlab_resource + action
        method = f"do_{self.gitlab_resource}_{self.resource_action}"
        if hasattr(self, method):
            return getattr(self, method)()

        # Fallback to standard actions (get, list, create, ...)
        method = f"do_{self.resource_action}"
        if hasattr(self, method):
            return getattr(self, method)()

        # Finally try to find custom methods
        return self.do_custom()

    def do_custom(self) -> Any:
        class_instance: (
            gitlab.base.RESTManager[gitlab.base.RESTObject] | gitlab.base.RESTObject
        )
        in_obj = cli.custom_actions[self.cls_name][self.resource_action].in_object

        # Get the object (lazy), then act
        if in_obj:
            data = {}
            if self.mgr._from_parent_attrs:
                for k in self.mgr._from_parent_attrs:
                    data[k] = self.parent_args[k]
            if not issubclass(self.cls, gitlab.mixins.GetWithoutIdMixin):
                if TYPE_CHECKING:
                    assert isinstance(self.cls._id_attr, str)
                data[self.cls._id_attr] = self.args.pop(self.cls._id_attr)
            class_instance = self.cls(self.mgr, data)
        else:
            class_instance = self.mgr

        method_name = self.resource_action.replace("-", "_")
        return getattr(class_instance, method_name)(**self.args)

    def do_project_export_download(self) -> None:
        try:
            project = self.gl.projects.get(self.parent_args["project_id"], lazy=True)
            export_status = project.exports.get()
            if TYPE_CHECKING:
                assert export_status is not None
            data = export_status.download()
            if TYPE_CHECKING:
                assert data is not None
                assert isinstance(data, bytes)
            sys.stdout.buffer.write(data)

        except Exception as e:  # pragma: no cover, cli.die is unit-tested
            cli.die("Impossible to download the export", e)

    def do_validate(self) -> None:
        if TYPE_CHECKING:
            assert isinstance(self.mgr, gitlab.v4.objects.CiLintManager)
        try:
            self.mgr.validate(self.args)
        except GitlabCiLintError as e:  # pragma: no cover, cli.die is unit-tested
            cli.die("CI YAML Lint failed", e)
        except Exception as e:  # pragma: no cover, cli.die is unit-tested
            cli.die("Cannot validate CI YAML", e)

    def do_create(self) -> gitlab.base.RESTObject:
        if TYPE_CHECKING:
            assert isinstance(self.mgr, gitlab.mixins.CreateMixin)
        try:
            result = self.mgr.create(self.args)
            if TYPE_CHECKING:
                assert isinstance(result, gitlab.base.RESTObject)
        except Exception as e:  # pragma: no cover, cli.die is unit-tested
            cli.die("Impossible to create object", e)
        return result

    def do_list(self) -> list[gitlab.base.RESTObject]:
        if TYPE_CHECKING:
            assert isinstance(self.mgr, gitlab.mixins.ListMixin)
        message_details = gitlab.utils.WarnMessageData(
            message=(
                "Your query returned {len_items} of {total_items} items. To return all "
                "items use `--get-all`. To silence this warning use `--no-get-all`."
            ),
            show_caller=False,
        )

        try:
            result = self.mgr.list(
                **self.args, message_details=message_details, iterator=False
            )
        except Exception as e:  # pragma: no cover, cli.die is unit-tested
            cli.die("Impossible to list objects", e)
        return result

    def do_get(self) -> gitlab.base.RESTObject | None:
        if isinstance(self.mgr, gitlab.mixins.GetWithoutIdMixin):
            try:
                result = self.mgr.get(id=None, **self.args)
                if TYPE_CHECKING:
                    assert isinstance(result, gitlab.base.RESTObject) or result is None
            except Exception as e:  # pragma: no cover, cli.die is unit-tested
                cli.die("Impossible to get object", e)
            return result

        if TYPE_CHECKING:
            assert isinstance(self.mgr, gitlab.mixins.GetMixin)
            assert isinstance(self.cls._id_attr, str)

        id = self.args.pop(self.cls._id_attr)
        try:
            result = self.mgr.get(id, lazy=False, **self.args)
            if TYPE_CHECKING:
                assert isinstance(result, gitlab.base.RESTObject) or result is None
        except Exception as e:  # pragma: no cover, cli.die is unit-tested
            cli.die("Impossible to get object", e)
        return result

    def do_delete(self) -> None:
        if TYPE_CHECKING:
            assert isinstance(self.mgr, gitlab.mixins.DeleteMixin)
            assert isinstance(self.cls._id_attr, str)
        id = self.args.pop(self.cls._id_attr)
        try:
            self.mgr.delete(id, **self.args)
        except Exception as e:  # pragma: no cover, cli.die is unit-tested
            cli.die("Impossible to destroy object", e)

    def do_update(self) -> dict[str, Any]:
        if TYPE_CHECKING:
            assert isinstance(self.mgr, gitlab.mixins.UpdateMixin)
        if issubclass(self.mgr_cls, gitlab.mixins.GetWithoutIdMixin):
            id = None
        else:
            if TYPE_CHECKING:
                assert isinstance(self.cls._id_attr, str)
            id = self.args.pop(self.cls._id_attr)

        try:
            result = self.mgr.update(id, self.args)
        except Exception as e:  # pragma: no cover, cli.die is unit-tested
            cli.die("Impossible to update object", e)
        return result


# https://github.com/python/typeshed/issues/7539#issuecomment-1076581049
if TYPE_CHECKING:
    _SubparserType = argparse._SubParsersAction[argparse.ArgumentParser]
else:
    _SubparserType = Any


def _populate_sub_parser_by_class(
    cls: type[gitlab.base.RESTObject], sub_parser: _SubparserType
) -> None:
    mgr_cls_name = f"{cls.__name__}Manager"
    mgr_cls = getattr(gitlab.v4.objects, mgr_cls_name)

    action_parsers: dict[str, argparse.ArgumentParser] = {}
    for action_name, help_text in [
        ("list", "List the GitLab resources"),
        ("get", "Get a GitLab resource"),
        ("create", "Create a GitLab resource"),
        ("update", "Update a GitLab resource"),
        ("delete", "Delete a GitLab resource"),
    ]:
        if not hasattr(mgr_cls, action_name):
            continue

        sub_parser_action = sub_parser.add_parser(
            action_name, conflict_handler="resolve", help=help_text
        )
        action_parsers[action_name] = sub_parser_action
        sub_parser_action.add_argument("--sudo", required=False)
        if mgr_cls._from_parent_attrs:
            for x in mgr_cls._from_parent_attrs:
                sub_parser_action.add_argument(
                    f"--{x.replace('_', '-')}", required=True
                )

        if action_name == "list":
            for x in mgr_cls._list_filters:
                sub_parser_action.add_argument(
                    f"--{x.replace('_', '-')}", required=False
                )

            sub_parser_action.add_argument("--page", required=False, type=int)
            sub_parser_action.add_argument("--per-page", required=False, type=int)
            get_all_group = sub_parser_action.add_mutually_exclusive_group()
            get_all_group.add_argument(
                "--get-all",
                required=False,
                action="store_const",
                const=True,
                default=None,
                dest="get_all",
                help="Return all items from the server, without pagination.",
            )
            get_all_group.add_argument(
                "--no-get-all",
                required=False,
                action="store_const",
                const=False,
                default=None,
                dest="get_all",
                help="Don't return all items from the server.",
            )

        if action_name == "delete":
            if cls._id_attr is not None:
                id_attr = cls._id_attr.replace("_", "-")
                sub_parser_action.add_argument(f"--{id_attr}", required=True)

        if action_name == "get":
            if not issubclass(cls, gitlab.mixins.GetWithoutIdMixin):
                if cls._id_attr is not None:
                    id_attr = cls._id_attr.replace("_", "-")
                    sub_parser_action.add_argument(f"--{id_attr}", required=True)

            for x in mgr_cls._optional_get_attrs:
                sub_parser_action.add_argument(
                    f"--{x.replace('_', '-')}", required=False
                )

        if action_name == "create":
            for x in mgr_cls._create_attrs.required:
                sub_parser_action.add_argument(
                    f"--{x.replace('_', '-')}", required=True
                )
            for x in mgr_cls._create_attrs.optional:
                sub_parser_action.add_argument(
                    f"--{x.replace('_', '-')}", required=False
                )
            if mgr_cls._create_attrs.exclusive:
                group = sub_parser_action.add_mutually_exclusive_group()
                for x in mgr_cls._create_attrs.exclusive:
                    group.add_argument(f"--{x.replace('_', '-')}")

        if action_name == "update":
            if cls._id_attr is not None:
                id_attr = cls._id_attr.replace("_", "-")
                sub_parser_action.add_argument(f"--{id_attr}", required=True)

            for x in mgr_cls._update_attrs.required:
                if x != cls._id_attr:
                    sub_parser_action.add_argument(
                        f"--{x.replace('_', '-')}", required=True
                    )

            for x in mgr_cls._update_attrs.optional:
                if x != cls._id_attr:
                    sub_parser_action.add_argument(
                        f"--{x.replace('_', '-')}", required=False
                    )

            if mgr_cls._update_attrs.exclusive:
                group = sub_parser_action.add_mutually_exclusive_group()
                for x in mgr_cls._update_attrs.exclusive:
                    group.add_argument(f"--{x.replace('_', '-')}")

    if cls.__name__ in cli.custom_actions:
        name = cls.__name__
        for action_name in cli.custom_actions[name]:
            custom_action = cli.custom_actions[name][action_name]
            # NOTE(jlvillal): If we put a function for the `default` value of
            # the `get` it will always get called, which will break things.
            action_parser = action_parsers.get(action_name)
            if action_parser is None:
                sub_parser_action = sub_parser.add_parser(
                    action_name, help=custom_action.help
                )
            else:
                sub_parser_action = action_parser
            # Get the attributes for URL/path construction
            if mgr_cls._from_parent_attrs:
                for x in mgr_cls._from_parent_attrs:
                    sub_parser_action.add_argument(
                        f"--{x.replace('_', '-')}", required=True
                    )
                sub_parser_action.add_argument("--sudo", required=False)

            # We need to get the object somehow
            if not issubclass(cls, gitlab.mixins.GetWithoutIdMixin):
                if cls._id_attr is not None and custom_action.requires_id:
                    id_attr = cls._id_attr.replace("_", "-")
                    sub_parser_action.add_argument(f"--{id_attr}", required=True)

            for x in custom_action.required:
                if x != cls._id_attr:
                    sub_parser_action.add_argument(
                        f"--{x.replace('_', '-')}", required=True
                    )
            for x in custom_action.optional:
                if x != cls._id_attr:
                    sub_parser_action.add_argument(
                        f"--{x.replace('_', '-')}", required=False
                    )

    if mgr_cls.__name__ in cli.custom_actions:
        name = mgr_cls.__name__
        for action_name in cli.custom_actions[name]:
            # NOTE(jlvillal): If we put a function for the `default` value of
            # the `get` it will always get called, which will break things.
            action_parser = action_parsers.get(action_name)
            if action_parser is None:
                sub_parser_action = sub_parser.add_parser(action_name)
            else:
                sub_parser_action = action_parser
            if mgr_cls._from_parent_attrs:
                for x in mgr_cls._from_parent_attrs:
                    sub_parser_action.add_argument(
                        f"--{x.replace('_', '-')}", required=True
                    )
                sub_parser_action.add_argument("--sudo", required=False)

            custom_action = cli.custom_actions[name][action_name]
            for x in custom_action.required:
                if x != cls._id_attr:
                    sub_parser_action.add_argument(
                        f"--{x.replace('_', '-')}", required=True
                    )
            for x in custom_action.optional:
                if x != cls._id_attr:
                    sub_parser_action.add_argument(
                        f"--{x.replace('_', '-')}", required=False
                    )


def extend_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    subparsers = parser.add_subparsers(
        title="resource",
        dest="gitlab_resource",
        help="The GitLab resource to manipulate.",
    )
    subparsers.required = True

    # populate argparse for all Gitlab Object
    classes: set[type[gitlab.base.RESTObject]] = set()
    for cls in gitlab.v4.objects.__dict__.values():
        if not isinstance(cls, type):
            continue
        if issubclass(cls, gitlab.base.RESTManager):
            classes.add(cls._obj_cls)

    for cls in sorted(classes, key=operator.attrgetter("__name__")):
        if cls is gitlab.base.RESTObject:
            # Skip managers where _obj_cls is a plain RESTObject class
            # Those managers do not actually manage any objects and
            # can only be used to calls specific API paths.
            continue

        arg_name = cli.cls_to_gitlab_resource(cls)
        mgr_cls_name = f"{cls.__name__}Manager"
        mgr_cls = getattr(gitlab.v4.objects, mgr_cls_name)
        object_group = subparsers.add_parser(
            arg_name, help=f"API endpoint: {mgr_cls._path}"
        )

        object_subparsers = object_group.add_subparsers(
            title="action",
            dest="resource_action",
            help="Action to execute on the GitLab resource.",
        )
        _populate_sub_parser_by_class(cls, object_subparsers)
        object_subparsers.required = True

    return parser


def get_dict(
    obj: str | dict[str, Any] | gitlab.base.RESTObject, fields: list[str]
) -> str | dict[str, Any]:
    if not isinstance(obj, gitlab.base.RESTObject):
        return obj

    if fields:
        return {k: v for k, v in obj.attributes.items() if k in fields}
    return obj.attributes


class JSONPrinter:
    @staticmethod
    def display(d: str | dict[str, Any], **_kwargs: Any) -> None:
        print(json.dumps(d))

    @staticmethod
    def display_list(
        data: list[str | dict[str, Any] | gitlab.base.RESTObject],
        fields: list[str],
        **_kwargs: Any,
    ) -> None:
        print(json.dumps([get_dict(obj, fields) for obj in data]))


class YAMLPrinter:
    @staticmethod
    def display(d: str | dict[str, Any], **_kwargs: Any) -> None:
        try:
            import yaml  # noqa

            print(yaml.safe_dump(d, default_flow_style=False))
        except ImportError:
            sys.exit(
                "PyYaml is not installed.\n"
                "Install it with `pip install PyYaml` "
                "to use the yaml output feature"
            )

    @staticmethod
    def display_list(
        data: list[str | dict[str, Any] | gitlab.base.RESTObject],
        fields: list[str],
        **_kwargs: Any,
    ) -> None:
        try:
            import yaml  # noqa

            print(
                yaml.safe_dump(
                    [get_dict(obj, fields) for obj in data], default_flow_style=False
                )
            )
        except ImportError:
            sys.exit(
                "PyYaml is not installed.\n"
                "Install it with `pip install PyYaml` "
                "to use the yaml output feature"
            )


class LegacyPrinter:
    def display(self, _d: str | dict[str, Any], **kwargs: Any) -> None:
        verbose = kwargs.get("verbose", False)
        padding = kwargs.get("padding", 0)
        obj: dict[str, Any] | gitlab.base.RESTObject | None = kwargs.get("obj")
        if TYPE_CHECKING:
            assert obj is not None

        def display_dict(d: dict[str, Any], padding: int) -> None:
            for k in sorted(d.keys()):
                v = d[k]
                if isinstance(v, dict):
                    print(f"{' ' * padding}{k.replace('_', '-')}:")
                    new_padding = padding + 2
                    self.display(v, verbose=True, padding=new_padding, obj=v)
                    continue
                print(f"{' ' * padding}{k.replace('_', '-')}: {v}")

        if verbose:
            if isinstance(obj, dict):
                display_dict(obj, padding)
                return

            # not a dict, we assume it's a RESTObject
            if obj._id_attr:
                id = getattr(obj, obj._id_attr, None)
                print(f"{obj._id_attr}: {id}")
            attrs = obj.attributes
            if obj._id_attr:
                attrs.pop(obj._id_attr)
            display_dict(attrs, padding)
            return

        lines = []

        if TYPE_CHECKING:
            assert isinstance(obj, gitlab.base.RESTObject)

        if obj._id_attr:
            id = getattr(obj, obj._id_attr)
            lines.append(f"{obj._id_attr.replace('_', '-')}: {id}")
        if obj._repr_attr:
            value = getattr(obj, obj._repr_attr, "None") or "None"
            value = value.replace("\r", "").replace("\n", " ")
            # If the attribute is a note (ProjectCommitComment) then we do
            # some modifications to fit everything on one line
            line = f"{obj._repr_attr}: {value}"
            # ellipsize long lines (comments)
            if len(line) > 79:
                line = f"{line[:76]}..."
            lines.append(line)

        if lines:
            print("\n".join(lines))
            return

        print(
            f"No default fields to show for {obj!r}. "
            f"Please use  '--verbose' or the JSON/YAML formatters."
        )

    def display_list(
        self, data: list[str | gitlab.base.RESTObject], fields: list[str], **kwargs: Any
    ) -> None:
        verbose = kwargs.get("verbose", False)
        for obj in data:
            if isinstance(obj, gitlab.base.RESTObject):
                self.display(get_dict(obj, fields), verbose=verbose, obj=obj)
            else:
                print(obj)
            print("")


PRINTERS: dict[str, type[JSONPrinter] | type[LegacyPrinter] | type[YAMLPrinter]] = {
    "json": JSONPrinter,
    "legacy": LegacyPrinter,
    "yaml": YAMLPrinter,
}


def run(
    gl: gitlab.Gitlab,
    gitlab_resource: str,
    resource_action: str,
    args: dict[str, Any],
    verbose: bool,
    output: str,
    fields: list[str],
) -> None:
    g_cli = GitlabCLI(
        gl=gl,
        gitlab_resource=gitlab_resource,
        resource_action=resource_action,
        args=args,
    )
    data = g_cli.run()

    printer: JSONPrinter | LegacyPrinter | YAMLPrinter = PRINTERS[output]()

    if isinstance(data, dict):
        printer.display(data, verbose=True, obj=data)
    elif isinstance(data, list):
        printer.display_list(data, fields, verbose=verbose)
    elif isinstance(data, gitlab.base.RESTObjectList):
        printer.display_list(list(data), fields, verbose=verbose)
    elif isinstance(data, gitlab.base.RESTObject):
        printer.display(get_dict(data, fields), verbose=verbose, obj=data)
    elif isinstance(data, str):
        print(data)
    elif isinstance(data, bytes):
        sys.stdout.buffer.write(data)
    elif hasattr(data, "decode"):
        print(data.decode())



from __future__ import annotations

from typing import Any

import requests

from gitlab import cli
from gitlab import exceptions as exc
from gitlab.base import RESTObject
from gitlab.mixins import (
    CreateMixin,
    DeleteMixin,
    ObjectDeleteMixin,
    RetrieveMixin,
    SaveMixin,
    UpdateMixin,
)
from gitlab.types import ArrayAttribute, RequiredOptional

__all__ = [
    "ProjectEnvironment",
    "ProjectEnvironmentManager",
    "ProjectProtectedEnvironment",
    "ProjectProtectedEnvironmentManager",
]


class ProjectEnvironment(SaveMixin, ObjectDeleteMixin, RESTObject):
    @cli.register_custom_action(cls_names="ProjectEnvironment")
    @exc.on_http_error(exc.GitlabStopError)
    def stop(self, **kwargs: Any) -> dict[str, Any] | requests.Response:
        """Stop the environment.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabStopError: If the operation failed

        Returns:
           A dict of the result.
        """
        path = f"{self.manager.path}/{self.encoded_id}/stop"
        return self.manager.gitlab.http_post(path, **kwargs)


class ProjectEnvironmentManager(
    RetrieveMixin[ProjectEnvironment],
    CreateMixin[ProjectEnvironment],
    UpdateMixin[ProjectEnvironment],
    DeleteMixin[ProjectEnvironment],
):
    _path = "/projects/{project_id}/environments"
    _obj_cls = ProjectEnvironment
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(required=("name",), optional=("external_url",))
    _update_attrs = RequiredOptional(optional=("name", "external_url"))
    _list_filters = ("name", "search", "states")


class ProjectProtectedEnvironment(ObjectDeleteMixin, RESTObject):
    _id_attr = "name"
    _repr_attr = "name"


class ProjectProtectedEnvironmentManager(
    RetrieveMixin[ProjectProtectedEnvironment],
    CreateMixin[ProjectProtectedEnvironment],
    DeleteMixin[ProjectProtectedEnvironment],
):
    _path = "/projects/{project_id}/protected_environments"
    _obj_cls = ProjectProtectedEnvironment
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("name", "deploy_access_levels"),
        optional=("required_approval_count", "approval_rules"),
    )
    _types = {"deploy_access_levels": ArrayAttribute, "approval_rules": ArrayAttribute}



from __future__ import annotations

from typing import Any

from gitlab import cli
from gitlab import exceptions as exc
from gitlab.base import RESTObject
from gitlab.mixins import (
    DeleteMixin,
    GetMixin,
    ListMixin,
    ObjectDeleteMixin,
    RetrieveMixin,
)

__all__ = [
    "GroupRegistryRepositoryManager",
    "ProjectRegistryRepository",
    "ProjectRegistryRepositoryManager",
    "ProjectRegistryTag",
    "ProjectRegistryTagManager",
    "RegistryRepository",
    "RegistryRepositoryManager",
]


class ProjectRegistryRepository(ObjectDeleteMixin, RESTObject):
    tags: ProjectRegistryTagManager


class ProjectRegistryRepositoryManager(
    DeleteMixin[ProjectRegistryRepository], ListMixin[ProjectRegistryRepository]
):
    _path = "/projects/{project_id}/registry/repositories"
    _obj_cls = ProjectRegistryRepository
    _from_parent_attrs = {"project_id": "id"}


class ProjectRegistryTag(ObjectDeleteMixin, RESTObject):
    _id_attr = "name"


class ProjectRegistryTagManager(
    DeleteMixin[ProjectRegistryTag], RetrieveMixin[ProjectRegistryTag]
):
    _obj_cls = ProjectRegistryTag
    _from_parent_attrs = {"project_id": "project_id", "repository_id": "id"}
    _path = "/projects/{project_id}/registry/repositories/{repository_id}/tags"

    @cli.register_custom_action(
        cls_names="ProjectRegistryTagManager",
        required=("name_regex_delete",),
        optional=("keep_n", "name_regex_keep", "older_than"),
    )
    @exc.on_http_error(exc.GitlabDeleteError)
    def delete_in_bulk(self, name_regex_delete: str, **kwargs: Any) -> None:
        """Delete Tag in bulk

        Args:
            name_regex_delete: The regex of the name to delete. To delete all
                tags specify .*.
            keep_n: The amount of latest tags of given name to keep.
            name_regex_keep: The regex of the name to keep. This value
                overrides any matches from name_regex.
            older_than: Tags to delete that are older than the given time,
                written in human readable form 1h, 1d, 1month.
            **kwargs: Extra options to send to the server (e.g. sudo)
        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabDeleteError: If the server cannot perform the request
        """
        valid_attrs = ["keep_n", "name_regex_keep", "older_than"]
        data = {"name_regex_delete": name_regex_delete}
        data.update({k: v for k, v in kwargs.items() if k in valid_attrs})
        self.gitlab.http_delete(self.path, query_data=data, **kwargs)


class GroupRegistryRepositoryManager(ListMixin[ProjectRegistryRepository]):
    _path = "/groups/{group_id}/registry/repositories"
    _obj_cls = ProjectRegistryRepository
    _from_parent_attrs = {"group_id": "id"}


class RegistryRepository(RESTObject):
    _repr_attr = "path"


class RegistryRepositoryManager(GetMixin[RegistryRepository]):
    _path = "/registry/repositories"
    _obj_cls = RegistryRepository



"""
GitLab API:
https://docs.gitlab.com/ee/api/users.html
https://docs.gitlab.com/ee/api/projects.html#list-projects-starred-by-a-user
"""

from __future__ import annotations

from typing import Any, cast, Literal, Optional, overload

import requests

from gitlab import cli
from gitlab import exceptions as exc
from gitlab import types
from gitlab.base import RESTObject, RESTObjectList
from gitlab.mixins import (
    CreateMixin,
    CRUDMixin,
    DeleteMixin,
    GetWithoutIdMixin,
    ListMixin,
    NoUpdateMixin,
    ObjectDeleteMixin,
    RetrieveMixin,
    SaveMixin,
    UpdateMixin,
)
from gitlab.types import ArrayAttribute, RequiredOptional

from .custom_attributes import UserCustomAttributeManager  # noqa: F401
from .events import UserEventManager  # noqa: F401
from .personal_access_tokens import UserPersonalAccessTokenManager  # noqa: F401

__all__ = [
    "CurrentUserEmail",
    "CurrentUserEmailManager",
    "CurrentUserGPGKey",
    "CurrentUserGPGKeyManager",
    "CurrentUserKey",
    "CurrentUserKeyManager",
    "CurrentUserRunner",
    "CurrentUserRunnerManager",
    "CurrentUserStatus",
    "CurrentUserStatusManager",
    "CurrentUser",
    "CurrentUserManager",
    "User",
    "UserManager",
    "ProjectUser",
    "ProjectUserManager",
    "StarredProject",
    "StarredProjectManager",
    "UserEmail",
    "UserEmailManager",
    "UserActivities",
    "UserStatus",
    "UserStatusManager",
    "UserActivitiesManager",
    "UserGPGKey",
    "UserGPGKeyManager",
    "UserKey",
    "UserKeyManager",
    "UserIdentityProviderManager",
    "UserImpersonationToken",
    "UserImpersonationTokenManager",
    "UserMembership",
    "UserMembershipManager",
    "UserProject",
    "UserProjectManager",
    "UserContributedProject",
    "UserContributedProjectManager",
]


class CurrentUserEmail(ObjectDeleteMixin, RESTObject):
    _repr_attr = "email"


class CurrentUserEmailManager(
    RetrieveMixin[CurrentUserEmail],
    CreateMixin[CurrentUserEmail],
    DeleteMixin[CurrentUserEmail],
):
    _path = "/user/emails"
    _obj_cls = CurrentUserEmail
    _create_attrs = RequiredOptional(required=("email",))


class CurrentUserGPGKey(ObjectDeleteMixin, RESTObject):
    pass


class CurrentUserGPGKeyManager(
    RetrieveMixin[CurrentUserGPGKey],
    CreateMixin[CurrentUserGPGKey],
    DeleteMixin[CurrentUserGPGKey],
):
    _path = "/user/gpg_keys"
    _obj_cls = CurrentUserGPGKey
    _create_attrs = RequiredOptional(required=("key",))


class CurrentUserKey(ObjectDeleteMixin, RESTObject):
    _repr_attr = "title"


class CurrentUserKeyManager(
    RetrieveMixin[CurrentUserKey],
    CreateMixin[CurrentUserKey],
    DeleteMixin[CurrentUserKey],
):
    _path = "/user/keys"
    _obj_cls = CurrentUserKey
    _create_attrs = RequiredOptional(required=("title", "key"))


class CurrentUserRunner(RESTObject):
    pass


class CurrentUserRunnerManager(CreateMixin[CurrentUserRunner]):
    _path = "/user/runners"
    _obj_cls = CurrentUserRunner
    _types = {"tag_list": types.CommaSeparatedListAttribute}
    _create_attrs = RequiredOptional(
        required=("runner_type",),
        optional=(
            "group_id",
            "project_id",
            "description",
            "paused",
            "locked",
            "run_untagged",
            "tag_list",
            "access_level",
            "maximum_timeout",
            "maintenance_note",
        ),
    )


class CurrentUserStatus(SaveMixin, RESTObject):
    _id_attr = None
    _repr_attr = "message"


class CurrentUserStatusManager(
    GetWithoutIdMixin[CurrentUserStatus], UpdateMixin[CurrentUserStatus]
):
    _path = "/user/status"
    _obj_cls = CurrentUserStatus
    _update_attrs = RequiredOptional(optional=("emoji", "message"))


class CurrentUser(RESTObject):
    _id_attr = None
    _repr_attr = "username"

    emails: CurrentUserEmailManager
    gpgkeys: CurrentUserGPGKeyManager
    keys: CurrentUserKeyManager
    runners: CurrentUserRunnerManager
    status: CurrentUserStatusManager


class CurrentUserManager(GetWithoutIdMixin[CurrentUser]):
    _path = "/user"
    _obj_cls = CurrentUser


class User(SaveMixin, ObjectDeleteMixin, RESTObject):
    _repr_attr = "username"

    customattributes: UserCustomAttributeManager
    emails: UserEmailManager
    events: UserEventManager
    followers_users: UserFollowersManager
    following_users: UserFollowingManager
    gpgkeys: UserGPGKeyManager
    identityproviders: UserIdentityProviderManager
    impersonationtokens: UserImpersonationTokenManager
    keys: UserKeyManager
    memberships: UserMembershipManager
    personal_access_tokens: UserPersonalAccessTokenManager
    projects: UserProjectManager
    contributed_projects: UserContributedProjectManager
    starred_projects: StarredProjectManager
    status: UserStatusManager

    @cli.register_custom_action(cls_names="User")
    @exc.on_http_error(exc.GitlabBlockError)
    def block(self, **kwargs: Any) -> bool | None:
        """Block the user.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabBlockError: If the user could not be blocked

        Returns:
            Whether the user status has been changed
        """
        path = f"/users/{self.encoded_id}/block"
        # NOTE: Undocumented behavior of the GitLab API is that it returns a
        # boolean or None
        server_data = cast(
            Optional[bool], self.manager.gitlab.http_post(path, **kwargs)
        )
        if server_data is True:
            self._attrs["state"] = "blocked"
        return server_data

    @cli.register_custom_action(cls_names="User")
    @exc.on_http_error(exc.GitlabFollowError)
    def follow(self, **kwargs: Any) -> dict[str, Any] | requests.Response:
        """Follow the user.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabFollowError: If the user could not be followed

        Returns:
            The new object data (*not* a RESTObject)
        """
        path = f"/users/{self.encoded_id}/follow"
        return self.manager.gitlab.http_post(path, **kwargs)

    @cli.register_custom_action(cls_names="User")
    @exc.on_http_error(exc.GitlabUnfollowError)
    def unfollow(self, **kwargs: Any) -> dict[str, Any] | requests.Response:
        """Unfollow the user.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabUnfollowError: If the user could not be followed

        Returns:
            The new object data (*not* a RESTObject)
        """
        path = f"/users/{self.encoded_id}/unfollow"
        return self.manager.gitlab.http_post(path, **kwargs)

    @cli.register_custom_action(cls_names="User")
    @exc.on_http_error(exc.GitlabUnblockError)
    def unblock(self, **kwargs: Any) -> bool | None:
        """Unblock the user.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabUnblockError: If the user could not be unblocked

        Returns:
            Whether the user status has been changed
        """
        path = f"/users/{self.encoded_id}/unblock"
        # NOTE: Undocumented behavior of the GitLab API is that it returns a
        # boolean or None
        server_data = cast(
            Optional[bool], self.manager.gitlab.http_post(path, **kwargs)
        )
        if server_data is True:
            self._attrs["state"] = "active"
        return server_data

    @cli.register_custom_action(cls_names="User")
    @exc.on_http_error(exc.GitlabDeactivateError)
    def deactivate(self, **kwargs: Any) -> dict[str, Any] | requests.Response:
        """Deactivate the user.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabDeactivateError: If the user could not be deactivated

        Returns:
            Whether the user status has been changed
        """
        path = f"/users/{self.encoded_id}/deactivate"
        server_data = self.manager.gitlab.http_post(path, **kwargs)
        if server_data:
            self._attrs["state"] = "deactivated"
        return server_data

    @cli.register_custom_action(cls_names="User")
    @exc.on_http_error(exc.GitlabActivateError)
    def activate(self, **kwargs: Any) -> dict[str, Any] | requests.Response:
        """Activate the user.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabActivateError: If the user could not be activated

        Returns:
            Whether the user status has been changed
        """
        path = f"/users/{self.encoded_id}/activate"
        server_data = self.manager.gitlab.http_post(path, **kwargs)
        if server_data:
            self._attrs["state"] = "active"
        return server_data

    @cli.register_custom_action(cls_names="User")
    @exc.on_http_error(exc.GitlabUserApproveError)
    def approve(self, **kwargs: Any) -> dict[str, Any] | requests.Response:
        """Approve a user creation request.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabUserApproveError: If the user could not be activated

        Returns:
            The new object data (*not* a RESTObject)
        """
        path = f"/users/{self.encoded_id}/approve"
        return self.manager.gitlab.http_post(path, **kwargs)

    @cli.register_custom_action(cls_names="User")
    @exc.on_http_error(exc.GitlabUserRejectError)
    def reject(self, **kwargs: Any) -> dict[str, Any] | requests.Response:
        """Reject a user creation request.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabUserRejectError: If the user could not be rejected

        Returns:
            The new object data (*not* a RESTObject)
        """
        path = f"/users/{self.encoded_id}/reject"
        return self.manager.gitlab.http_post(path, **kwargs)

    @cli.register_custom_action(cls_names="User")
    @exc.on_http_error(exc.GitlabBanError)
    def ban(self, **kwargs: Any) -> dict[str, Any] | requests.Response:
        """Ban the user.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabBanError: If the user could not be banned

        Returns:
            Whether the user has been banned
        """
        path = f"/users/{self.encoded_id}/ban"
        server_data = self.manager.gitlab.http_post(path, **kwargs)
        if server_data:
            self._attrs["state"] = "banned"
        return server_data

    @cli.register_custom_action(cls_names="User")
    @exc.on_http_error(exc.GitlabUnbanError)
    def unban(self, **kwargs: Any) -> dict[str, Any] | requests.Response:
        """Unban the user.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabUnbanError: If the user could not be unbanned

        Returns:
            Whether the user has been unbanned
        """
        path = f"/users/{self.encoded_id}/unban"
        server_data = self.manager.gitlab.http_post(path, **kwargs)
        if server_data:
            self._attrs["state"] = "active"
        return server_data


class UserManager(CRUDMixin[User]):
    _path = "/users"
    _obj_cls = User

    _list_filters = (
        "username",
        "public_email",
        "search",
        "active",
        "external",
        "blocked",
        "humans",
        "created_after",
        "created_before",
        "exclude_active",
        "exclude_external",
        "exclude_humans",
        "exclude_internal",
        "without_project_bots",
        "extern_uid",
        "provider",
        "two_factor",
        "without_projects",
        "admins",
        "auditors",
        "skip_ldap",
        "custom_attributes",
        "status",
    )
    _create_attrs = RequiredOptional(
        optional=(
            "email",
            "username",
            "name",
            "password",
            "reset_password",
            "skype",
            "linkedin",
            "twitter",
            "projects_limit",
            "extern_uid",
            "provider",
            "bio",
            "admin",
            "can_create_group",
            "website_url",
            "skip_confirmation",
            "external",
            "organization",
            "location",
            "avatar",
            "public_email",
            "private_profile",
            "color_scheme_id",
            "theme_id",
        )
    )
    _update_attrs = RequiredOptional(
        required=("email", "username", "name"),
        optional=(
            "password",
            "skype",
            "linkedin",
            "twitter",
            "projects_limit",
            "extern_uid",
            "provider",
            "bio",
            "admin",
            "can_create_group",
            "website_url",
            "skip_reconfirmation",
            "external",
            "organization",
            "location",
            "avatar",
            "public_email",
            "private_profile",
            "color_scheme_id",
            "theme_id",
        ),
    )
    _types = {"confirm": types.LowercaseStringAttribute, "avatar": types.ImageAttribute}


class ProjectUser(RESTObject):
    pass


class ProjectUserManager(ListMixin[ProjectUser]):
    _path = "/projects/{project_id}/users"
    _obj_cls = ProjectUser
    _from_parent_attrs = {"project_id": "id"}
    _list_filters = ("search", "skip_users")
    _types = {"skip_users": types.ArrayAttribute}


class UserEmail(ObjectDeleteMixin, RESTObject):
    _repr_attr = "email"


class UserEmailManager(
    RetrieveMixin[UserEmail], CreateMixin[UserEmail], DeleteMixin[UserEmail]
):
    _path = "/users/{user_id}/emails"
    _obj_cls = UserEmail
    _from_parent_attrs = {"user_id": "id"}
    _create_attrs = RequiredOptional(
        required=("email",), optional=("skip_confirmation",)
    )


class UserActivities(RESTObject):
    _id_attr = "username"


class UserStatus(RESTObject):
    _id_attr = None
    _repr_attr = "message"


class UserStatusManager(GetWithoutIdMixin[UserStatus]):
    _path = "/users/{user_id}/status"
    _obj_cls = UserStatus
    _from_parent_attrs = {"user_id": "id"}


class UserActivitiesManager(ListMixin[UserActivities]):
    _path = "/user/activities"
    _obj_cls = UserActivities


class UserGPGKey(ObjectDeleteMixin, RESTObject):
    pass


class UserGPGKeyManager(
    RetrieveMixin[UserGPGKey], CreateMixin[UserGPGKey], DeleteMixin[UserGPGKey]
):
    _path = "/users/{user_id}/gpg_keys"
    _obj_cls = UserGPGKey
    _from_parent_attrs = {"user_id": "id"}
    _create_attrs = RequiredOptional(required=("key",))


class UserKey(ObjectDeleteMixin, RESTObject):
    pass


class UserKeyManager(
    RetrieveMixin[UserKey], CreateMixin[UserKey], DeleteMixin[UserKey]
):
    _path = "/users/{user_id}/keys"
    _obj_cls = UserKey
    _from_parent_attrs = {"user_id": "id"}
    _create_attrs = RequiredOptional(required=("title", "key"))


class UserIdentityProviderManager(DeleteMixin[User]):
    """Manager for user identities.

    This manager does not actually manage objects but enables
    functionality for deletion of user identities by provider.
    """

    _path = "/users/{user_id}/identities"
    _obj_cls = User
    _from_parent_attrs = {"user_id": "id"}


class UserImpersonationToken(ObjectDeleteMixin, RESTObject):
    pass


class UserImpersonationTokenManager(NoUpdateMixin[UserImpersonationToken]):
    _path = "/users/{user_id}/impersonation_tokens"
    _obj_cls = UserImpersonationToken
    _from_parent_attrs = {"user_id": "id"}
    _create_attrs = RequiredOptional(
        required=("name", "scopes"), optional=("expires_at",)
    )
    _list_filters = ("state",)
    _types = {"scopes": ArrayAttribute}


class UserMembership(RESTObject):
    _id_attr = "source_id"


class UserMembershipManager(RetrieveMixin[UserMembership]):
    _path = "/users/{user_id}/memberships"
    _obj_cls = UserMembership
    _from_parent_attrs = {"user_id": "id"}
    _list_filters = ("type",)


# Having this outside projects avoids circular imports due to ProjectUser
class UserProject(RESTObject):
    pass


class UserProjectManager(ListMixin[UserProject], CreateMixin[UserProject]):
    _path = "/projects/user/{user_id}"
    _obj_cls = UserProject
    _from_parent_attrs = {"user_id": "id"}
    _create_attrs = RequiredOptional(
        required=("name",),
        optional=(
            "default_branch",
            "issues_enabled",
            "wall_enabled",
            "merge_requests_enabled",
            "wiki_enabled",
            "snippets_enabled",
            "squash_option",
            "public",
            "visibility",
            "description",
            "builds_enabled",
            "public_builds",
            "import_url",
            "only_allow_merge_if_build_succeeds",
        ),
    )
    _list_filters = (
        "archived",
        "visibility",
        "order_by",
        "sort",
        "search",
        "simple",
        "owned",
        "membership",
        "starred",
        "statistics",
        "with_issues_enabled",
        "with_merge_requests_enabled",
        "with_custom_attributes",
        "with_programming_language",
        "wiki_checksum_failed",
        "repository_checksum_failed",
        "min_access_level",
        "id_after",
        "id_before",
    )

    @overload
    def list(
        self, *, iterator: Literal[False] = False, **kwargs: Any
    ) -> list[UserProject]: ...

    @overload
    def list(
        self, *, iterator: Literal[True] = True, **kwargs: Any
    ) -> RESTObjectList[UserProject]: ...

    @overload
    def list(
        self, *, iterator: bool = False, **kwargs: Any
    ) -> RESTObjectList[UserProject] | list[UserProject]: ...

    def list(
        self, *, iterator: bool = False, **kwargs: Any
    ) -> RESTObjectList[UserProject] | list[UserProject]:
        """Retrieve a list of objects.

        Args:
            get_all: If True, return all the items, without pagination
            per_page: Number of items to retrieve per request
            page: ID of the page to return (starts with page 1)
            iterator: If set to True and no pagination option is
                defined, return a generator instead of a list
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            The list of objects, or a generator if `iterator` is True

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabListError: If the server cannot perform the request
        """
        if self._parent:
            path = f"/users/{self._parent.id}/projects"
        else:
            path = f"/users/{self._from_parent_attrs['user_id']}/projects"
        return super().list(path=path, iterator=iterator, **kwargs)


class UserContributedProject(RESTObject):
    _id_attr = "id"
    _repr_attr = "path_with_namespace"


class UserContributedProjectManager(ListMixin[UserContributedProject]):
    _path = "/users/{user_id}/contributed_projects"
    _obj_cls = UserContributedProject
    _from_parent_attrs = {"user_id": "id"}


class StarredProject(RESTObject):
    pass


class StarredProjectManager(ListMixin[StarredProject]):
    _path = "/users/{user_id}/starred_projects"
    _obj_cls = StarredProject
    _from_parent_attrs = {"user_id": "id"}
    _list_filters = (
        "archived",
        "membership",
        "min_access_level",
        "order_by",
        "owned",
        "search",
        "simple",
        "sort",
        "starred",
        "statistics",
        "visibility",
        "with_custom_attributes",
        "with_issues_enabled",
        "with_merge_requests_enabled",
    )


class UserFollowersManager(ListMixin[User]):
    _path = "/users/{user_id}/followers"
    _obj_cls = User
    _from_parent_attrs = {"user_id": "id"}


class UserFollowingManager(ListMixin[User]):
    _path = "/users/{user_id}/following"
    _obj_cls = User
    _from_parent_attrs = {"user_id": "id"}



from gitlab.base import RESTObject
from gitlab.mixins import NoUpdateMixin, ObjectDeleteMixin
from gitlab.types import RequiredOptional

__all__ = [
    "ProjectTag",
    "ProjectTagManager",
    "ProjectProtectedTag",
    "ProjectProtectedTagManager",
]


class ProjectTag(ObjectDeleteMixin, RESTObject):
    _id_attr = "name"
    _repr_attr = "name"


class ProjectTagManager(NoUpdateMixin[ProjectTag]):
    _path = "/projects/{project_id}/repository/tags"
    _obj_cls = ProjectTag
    _from_parent_attrs = {"project_id": "id"}
    _list_filters = ("order_by", "sort", "search")
    _create_attrs = RequiredOptional(
        required=("tag_name", "ref"), optional=("message",)
    )


class ProjectProtectedTag(ObjectDeleteMixin, RESTObject):
    _id_attr = "name"
    _repr_attr = "name"


class ProjectProtectedTagManager(NoUpdateMixin[ProjectProtectedTag]):
    _path = "/projects/{project_id}/protected_tags"
    _obj_cls = ProjectProtectedTag
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("name",), optional=("create_access_level",)
    )



from gitlab.base import RESTObject
from gitlab.mixins import (
    CreateMixin,
    DeleteMixin,
    ObjectDeleteMixin,
    ObjectRotateMixin,
    RetrieveMixin,
    RotateMixin,
)
from gitlab.types import ArrayAttribute, RequiredOptional

__all__ = ["GroupAccessToken", "GroupAccessTokenManager"]


class GroupAccessToken(ObjectDeleteMixin, ObjectRotateMixin, RESTObject):
    pass


class GroupAccessTokenManager(
    CreateMixin[GroupAccessToken],
    DeleteMixin[GroupAccessToken],
    RetrieveMixin[GroupAccessToken],
    RotateMixin[GroupAccessToken],
):
    _path = "/groups/{group_id}/access_tokens"
    _obj_cls = GroupAccessToken
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = RequiredOptional(
        required=("name", "scopes"), optional=("access_level", "expires_at")
    )
    _types = {"scopes": ArrayAttribute}



from gitlab import types
from gitlab.base import RESTObject
from gitlab.mixins import (
    CreateMixin,
    DeleteMixin,
    ListMixin,
    ObjectDeleteMixin,
    RetrieveMixin,
)
from gitlab.types import RequiredOptional

__all__ = [
    "DeployToken",
    "DeployTokenManager",
    "GroupDeployToken",
    "GroupDeployTokenManager",
    "ProjectDeployToken",
    "ProjectDeployTokenManager",
]


class DeployToken(ObjectDeleteMixin, RESTObject):
    pass


class DeployTokenManager(ListMixin[DeployToken]):
    _path = "/deploy_tokens"
    _obj_cls = DeployToken


class GroupDeployToken(ObjectDeleteMixin, RESTObject):
    pass


class GroupDeployTokenManager(
    RetrieveMixin[GroupDeployToken],
    CreateMixin[GroupDeployToken],
    DeleteMixin[GroupDeployToken],
):
    _path = "/groups/{group_id}/deploy_tokens"
    _from_parent_attrs = {"group_id": "id"}
    _obj_cls = GroupDeployToken
    _create_attrs = RequiredOptional(
        required=("name", "scopes"), optional=("expires_at", "username")
    )
    _list_filters = ("scopes",)
    _types = {"scopes": types.ArrayAttribute}


class ProjectDeployToken(ObjectDeleteMixin, RESTObject):
    pass


class ProjectDeployTokenManager(
    RetrieveMixin[ProjectDeployToken],
    CreateMixin[ProjectDeployToken],
    DeleteMixin[ProjectDeployToken],
):
    _path = "/projects/{project_id}/deploy_tokens"
    _from_parent_attrs = {"project_id": "id"}
    _obj_cls = ProjectDeployToken
    _create_attrs = RequiredOptional(
        required=("name", "scopes"), optional=("expires_at", "username")
    )
    _list_filters = ("scopes",)
    _types = {"scopes": types.ArrayAttribute}



from gitlab.base import RESTObject
from gitlab.mixins import CRUDMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import RequiredOptional

__all__ = [
    "GroupBoardList",
    "GroupBoardListManager",
    "GroupBoard",
    "GroupBoardManager",
    "ProjectBoardList",
    "ProjectBoardListManager",
    "ProjectBoard",
    "ProjectBoardManager",
]


class GroupBoardList(SaveMixin, ObjectDeleteMixin, RESTObject):
    pass


class GroupBoardListManager(CRUDMixin[GroupBoardList]):
    _path = "/groups/{group_id}/boards/{board_id}/lists"
    _obj_cls = GroupBoardList
    _from_parent_attrs = {"group_id": "group_id", "board_id": "id"}
    _create_attrs = RequiredOptional(
        exclusive=("label_id", "assignee_id", "milestone_id", "iteration_id")
    )
    _update_attrs = RequiredOptional(required=("position",))


class GroupBoard(SaveMixin, ObjectDeleteMixin, RESTObject):
    lists: GroupBoardListManager


class GroupBoardManager(CRUDMixin[GroupBoard]):
    _path = "/groups/{group_id}/boards"
    _obj_cls = GroupBoard
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = RequiredOptional(required=("name",))


class ProjectBoardList(SaveMixin, ObjectDeleteMixin, RESTObject):
    pass


class ProjectBoardListManager(CRUDMixin[ProjectBoardList]):
    _path = "/projects/{project_id}/boards/{board_id}/lists"
    _obj_cls = ProjectBoardList
    _from_parent_attrs = {"project_id": "project_id", "board_id": "id"}
    _create_attrs = RequiredOptional(
        exclusive=("label_id", "assignee_id", "milestone_id", "iteration_id")
    )
    _update_attrs = RequiredOptional(required=("position",))


class ProjectBoard(SaveMixin, ObjectDeleteMixin, RESTObject):
    lists: ProjectBoardListManager


class ProjectBoardManager(CRUDMixin[ProjectBoard]):
    _path = "/projects/{project_id}/boards"
    _obj_cls = ProjectBoard
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(required=("name",))



from __future__ import annotations

from typing import cast

from gitlab.base import RESTObject
from gitlab.mixins import (
    CreateMixin,
    DeleteMixin,
    GetWithoutIdMixin,
    ListMixin,
    ObjectDeleteMixin,
    RefreshMixin,
    SaveMixin,
    UpdateMethod,
    UpdateMixin,
)
from gitlab.types import RequiredOptional

__all__ = ["ProjectJobTokenScope", "ProjectJobTokenScopeManager"]


class ProjectJobTokenScope(RefreshMixin, SaveMixin, RESTObject):
    _id_attr = None

    allowlist: AllowlistProjectManager
    groups_allowlist: AllowlistGroupManager


class ProjectJobTokenScopeManager(
    GetWithoutIdMixin[ProjectJobTokenScope], UpdateMixin[ProjectJobTokenScope]
):
    _path = "/projects/{project_id}/job_token_scope"
    _obj_cls = ProjectJobTokenScope
    _from_parent_attrs = {"project_id": "id"}
    _update_method = UpdateMethod.PATCH


class AllowlistProject(ObjectDeleteMixin, RESTObject):
    _id_attr = "target_project_id"  # note: only true for create endpoint

    def get_id(self) -> int:
        """Returns the id of the resource. This override deals with
        the fact that either an `id` or a `target_project_id` attribute
        is returned by the server depending on the endpoint called."""
        target_project_id = cast(int, super().get_id())
        if target_project_id is not None:
            return target_project_id
        return cast(int, self.id)


class AllowlistProjectManager(
    ListMixin[AllowlistProject],
    CreateMixin[AllowlistProject],
    DeleteMixin[AllowlistProject],
):
    _path = "/projects/{project_id}/job_token_scope/allowlist"
    _obj_cls = AllowlistProject
    _from_parent_attrs = {"project_id": "project_id"}
    _create_attrs = RequiredOptional(required=("target_project_id",))


class AllowlistGroup(ObjectDeleteMixin, RESTObject):
    _id_attr = "target_group_id"  # note: only true for create endpoint

    def get_id(self) -> int:
        """Returns the id of the resource. This override deals with
        the fact that either an `id` or a `target_group_id` attribute
        is returned by the server depending on the endpoint called."""
        target_group_id = cast(int, super().get_id())
        if target_group_id is not None:
            return target_group_id
        return cast(int, self.id)


class AllowlistGroupManager(
    ListMixin[AllowlistGroup], CreateMixin[AllowlistGroup], DeleteMixin[AllowlistGroup]
):
    _path = "/projects/{project_id}/job_token_scope/groups_allowlist"
    _obj_cls = AllowlistGroup
    _from_parent_attrs = {"project_id": "project_id"}
    _create_attrs = RequiredOptional(required=("target_group_id",))



from __future__ import annotations

from typing import Any

from gitlab import exceptions as exc
from gitlab import types
from gitlab.base import RESTObject
from gitlab.mixins import GetWithoutIdMixin, SaveMixin, UpdateMixin
from gitlab.types import RequiredOptional

__all__ = ["ApplicationSettings", "ApplicationSettingsManager"]


class ApplicationSettings(SaveMixin, RESTObject):
    _id_attr = None


class ApplicationSettingsManager(
    GetWithoutIdMixin[ApplicationSettings], UpdateMixin[ApplicationSettings]
):
    _path = "/application/settings"
    _obj_cls = ApplicationSettings
    _update_attrs = RequiredOptional(
        optional=(
            "id",
            "default_projects_limit",
            "signup_enabled",
            "silent_mode_enabled",
            "password_authentication_enabled_for_web",
            "gravatar_enabled",
            "sign_in_text",
            "created_at",
            "updated_at",
            "home_page_url",
            "default_branch_protection",
            "restricted_visibility_levels",
            "max_attachment_size",
            "session_expire_delay",
            "default_project_visibility",
            "default_snippet_visibility",
            "default_group_visibility",
            "outbound_local_requests_whitelist",
            "disabled_oauth_sign_in_sources",
            "domain_whitelist",
            "domain_blacklist_enabled",
            "domain_blacklist",
            "domain_allowlist",
            "domain_denylist_enabled",
            "domain_denylist",
            "external_authorization_service_enabled",
            "external_authorization_service_url",
            "external_authorization_service_default_label",
            "external_authorization_service_timeout",
            "import_sources",
            "user_oauth_applications",
            "after_sign_out_path",
            "container_registry_token_expire_delay",
            "repository_storages",
            "plantuml_enabled",
            "plantuml_url",
            "terminal_max_session_time",
            "polling_interval_multiplier",
            "rsa_key_restriction",
            "dsa_key_restriction",
            "ecdsa_key_restriction",
            "ed25519_key_restriction",
            "first_day_of_week",
            "enforce_terms",
            "terms",
            "performance_bar_allowed_group_id",
            "instance_statistics_visibility_private",
            "user_show_add_ssh_key_message",
            "file_template_project_id",
            "local_markdown_version",
            "asset_proxy_enabled",
            "asset_proxy_url",
            "asset_proxy_whitelist",
            "asset_proxy_allowlist",
            "geo_node_allowed_ips",
            "allow_local_requests_from_hooks_and_services",
            "allow_local_requests_from_web_hooks_and_services",
            "allow_local_requests_from_system_hooks",
        )
    )
    _types = {
        "asset_proxy_allowlist": types.ArrayAttribute,
        "disabled_oauth_sign_in_sources": types.ArrayAttribute,
        "domain_allowlist": types.ArrayAttribute,
        "domain_denylist": types.ArrayAttribute,
        "import_sources": types.ArrayAttribute,
        "restricted_visibility_levels": types.ArrayAttribute,
    }

    @exc.on_http_error(exc.GitlabUpdateError)
    def update(
        self,
        id: str | int | None = None,
        new_data: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Update an object on the server.

        Args:
            id: ID of the object to update (can be None if not required)
            new_data: the update data for the object
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            The new object data (*not* a RESTObject)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabUpdateError: If the server cannot perform the request
        """
        new_data = new_data or {}
        data = new_data.copy()
        if "domain_whitelist" in data and data["domain_whitelist"] is None:
            data.pop("domain_whitelist")
        return super().update(id, data, **kwargs)



__author__ = "Gauvain Pocentek, python-gitlab team"
__copyright__ = "Copyright 2013-2019 Gauvain Pocentek, 2019-2023 python-gitlab team"
__email__ = "gauvainpocentek@gmail.com"
__license__ = "LGPL3"
__title__ = "python-gitlab"
__version__ = "8.3.0"



from gitlab.base import RESTObject
from gitlab.mixins import CRUDMixin, ObjectDeleteMixin, SaveMixin, UploadMixin
from gitlab.types import RequiredOptional

__all__ = ["ProjectWiki", "ProjectWikiManager", "GroupWiki", "GroupWikiManager"]


class ProjectWiki(SaveMixin, ObjectDeleteMixin, UploadMixin, RESTObject):
    _id_attr = "slug"
    _repr_attr = "slug"
    _upload_path = "/projects/{project_id}/wikis/attachments"


class ProjectWikiManager(CRUDMixin[ProjectWiki]):
    _path = "/projects/{project_id}/wikis"
    _obj_cls = ProjectWiki
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("title", "content"), optional=("format",)
    )
    _update_attrs = RequiredOptional(optional=("title", "content", "format"))
    _list_filters = ("with_content",)


class GroupWiki(SaveMixin, ObjectDeleteMixin, UploadMixin, RESTObject):
    _id_attr = "slug"
    _repr_attr = "slug"
    _upload_path = "/groups/{group_id}/wikis/attachments"


class GroupWikiManager(CRUDMixin[GroupWiki]):
    _path = "/groups/{group_id}/wikis"
    _obj_cls = GroupWiki
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = RequiredOptional(
        required=("title", "content"), optional=("format",)
    )
    _update_attrs = RequiredOptional(optional=("title", "content", "format"))
    _list_filters = ("with_content",)



from __future__ import annotations

import configparser
import os
import shlex
import subprocess
from os.path import expanduser, expandvars
from pathlib import Path

from gitlab.const import USER_AGENT

_DEFAULT_FILES: list[str] = [
    "/etc/python-gitlab.cfg",
    str(Path.home() / ".python-gitlab.cfg"),
]

HELPER_PREFIX = "helper:"

HELPER_ATTRIBUTES = ["job_token", "http_password", "private_token", "oauth_token"]

_CONFIG_PARSER_ERRORS = (configparser.NoOptionError, configparser.NoSectionError)


def _resolve_file(filepath: Path | str) -> str:
    resolved = Path(filepath).resolve(strict=True)
    return str(resolved)


def _get_config_files(config_files: list[str] | None = None) -> str | list[str]:
    """
    Return resolved path(s) to config files if they exist, with precedence:
    1. Files passed in config_files
    2. File defined in PYTHON_GITLAB_CFG
    3. User- and system-wide config files
    """
    resolved_files = []

    if config_files:
        for config_file in config_files:
            try:
                resolved = _resolve_file(config_file)
            except OSError as e:
                raise GitlabConfigMissingError(
                    f"Cannot read config from file: {e}"
                ) from e
            resolved_files.append(resolved)

        return resolved_files

    try:
        env_config = os.environ["PYTHON_GITLAB_CFG"]
        return _resolve_file(env_config)
    except KeyError:
        pass
    except OSError as e:
        raise GitlabConfigMissingError(
            f"Cannot read config from PYTHON_GITLAB_CFG: {e}"
        ) from e

    for config_file in _DEFAULT_FILES:
        try:
            resolved = _resolve_file(config_file)
        except OSError:
            continue
        resolved_files.append(resolved)

    return resolved_files


class ConfigError(Exception):
    pass


class GitlabIDError(ConfigError):
    pass


class GitlabDataError(ConfigError):
    pass


class GitlabConfigMissingError(ConfigError):
    pass


class GitlabConfigHelperError(ConfigError):
    pass


class GitlabConfigParser:
    def __init__(
        self, gitlab_id: str | None = None, config_files: list[str] | None = None
    ) -> None:
        self.gitlab_id = gitlab_id
        self.http_username: str | None = None
        self.http_password: str | None = None
        self.job_token: str | None = None
        self.oauth_token: str | None = None
        self.private_token: str | None = None

        self.api_version: str = "4"
        self.order_by: str | None = None
        self.pagination: str | None = None
        self.per_page: int | None = None
        self.retry_transient_errors: bool = False
        self.ssl_verify: bool | str = True
        self.timeout: int = 60
        self.url: str | None = None
        self.user_agent: str = USER_AGENT
        self.keep_base_url: bool = False

        self._files = _get_config_files(config_files)
        if self._files:
            self._parse_config()

        if self.gitlab_id and not self._files:
            raise GitlabConfigMissingError(
                f"A gitlab id was provided ({self.gitlab_id}) but no config file found"
            )

    def _parse_config(self) -> None:
        _config = configparser.ConfigParser()
        _config.read(self._files, encoding="utf-8")

        if self.gitlab_id and not _config.has_section(self.gitlab_id):
            raise GitlabDataError(
                f"A gitlab id was provided ({self.gitlab_id}) "
                "but no config section found"
            )

        if self.gitlab_id is None:
            try:
                self.gitlab_id = _config.get("global", "default")
            except Exception as e:
                raise GitlabIDError(
                    "Impossible to get the gitlab id (not specified in config file)"
                ) from e

        try:
            self.url = _config.get(self.gitlab_id, "url")
        except Exception as e:
            raise GitlabDataError(
                "Impossible to get gitlab details from "
                f"configuration ({self.gitlab_id})"
            ) from e

        try:
            self.ssl_verify = _config.getboolean("global", "ssl_verify")
        except ValueError:
            # Value Error means the option exists but isn't a boolean.
            # Get as a string instead as it should then be a local path to a
            # CA bundle.
            self.ssl_verify = _config.get("global", "ssl_verify")
        except _CONFIG_PARSER_ERRORS:
            pass
        try:
            self.ssl_verify = _config.getboolean(self.gitlab_id, "ssl_verify")
        except ValueError:
            # Value Error means the option exists but isn't a boolean.
            # Get as a string instead as it should then be a local path to a
            # CA bundle.
            self.ssl_verify = _config.get(self.gitlab_id, "ssl_verify")
        except _CONFIG_PARSER_ERRORS:
            pass

        try:
            self.timeout = _config.getint("global", "timeout")
        except _CONFIG_PARSER_ERRORS:
            pass
        try:
            self.timeout = _config.getint(self.gitlab_id, "timeout")
        except _CONFIG_PARSER_ERRORS:
            pass

        try:
            self.private_token = _config.get(self.gitlab_id, "private_token")
        except _CONFIG_PARSER_ERRORS:
            pass

        try:
            self.oauth_token = _config.get(self.gitlab_id, "oauth_token")
        except _CONFIG_PARSER_ERRORS:
            pass

        try:
            self.job_token = _config.get(self.gitlab_id, "job_token")
        except _CONFIG_PARSER_ERRORS:
            pass

        try:
            self.http_username = _config.get(self.gitlab_id, "http_username")
            self.http_password = _config.get(
                self.gitlab_id, "http_password"
            )  # pragma: no cover
        except _CONFIG_PARSER_ERRORS:
            pass

        self._get_values_from_helper()

        try:
            self.api_version = _config.get("global", "api_version")
        except _CONFIG_PARSER_ERRORS:
            pass
        try:
            self.api_version = _config.get(self.gitlab_id, "api_version")
        except _CONFIG_PARSER_ERRORS:
            pass
        if self.api_version not in ("4",):
            raise GitlabDataError(f"Unsupported API version: {self.api_version}")

        for section in ["global", self.gitlab_id]:
            try:
                self.per_page = _config.getint(section, "per_page")
            except _CONFIG_PARSER_ERRORS:
                pass
        if self.per_page is not None and not 0 <= self.per_page <= 100:
            raise GitlabDataError(f"Unsupported per_page number: {self.per_page}")

        try:
            self.pagination = _config.get(self.gitlab_id, "pagination")
        except _CONFIG_PARSER_ERRORS:
            pass

        try:
            self.order_by = _config.get(self.gitlab_id, "order_by")
        except _CONFIG_PARSER_ERRORS:
            pass

        try:
            self.user_agent = _config.get("global", "user_agent")
        except _CONFIG_PARSER_ERRORS:
            pass
        try:
            self.user_agent = _config.get(self.gitlab_id, "user_agent")
        except _CONFIG_PARSER_ERRORS:
            pass

        try:
            self.keep_base_url = _config.getboolean("global", "keep_base_url")
        except _CONFIG_PARSER_ERRORS:
            pass
        try:
            self.keep_base_url = _config.getboolean(self.gitlab_id, "keep_base_url")
        except _CONFIG_PARSER_ERRORS:
            pass

        try:
            self.retry_transient_errors = _config.getboolean(
                "global", "retry_transient_errors"
            )
        except _CONFIG_PARSER_ERRORS:
            pass
        try:
            self.retry_transient_errors = _config.getboolean(
                self.gitlab_id, "retry_transient_errors"
            )
        except _CONFIG_PARSER_ERRORS:
            pass

    def _get_values_from_helper(self) -> None:
        """Update attributes that may get values from an external helper program"""
        for attr in HELPER_ATTRIBUTES:
            value = getattr(self, attr)
            if not isinstance(value, str):
                continue

            if not value.lower().strip().startswith(HELPER_PREFIX):
                continue

            helper = value[len(HELPER_PREFIX) :].strip()
            commmand = [expanduser(expandvars(token)) for token in shlex.split(helper)]

            try:
                value = (
                    subprocess.check_output(commmand, stderr=subprocess.PIPE)
                    .decode("utf-8")
                    .strip()
                )
            except subprocess.CalledProcessError as e:
                stderr = e.stderr.decode().strip()
                raise GitlabConfigHelperError(
                    f"Failed to read {attr} value from helper "
                    f"for {self.gitlab_id}:\n{stderr}"
                ) from e

            setattr(self, attr, value)



from __future__ import annotations

import dataclasses
from typing import Any, BinaryIO, TYPE_CHECKING

import requests
from requests import PreparedRequest
from requests.auth import AuthBase
from requests.structures import CaseInsensitiveDict
from requests_toolbelt.multipart.encoder import MultipartEncoder  # type: ignore

from . import protocol


class TokenAuth:
    def __init__(self, token: str):
        self.token = token


class OAuthTokenAuth(TokenAuth, AuthBase):
    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        r.headers["Authorization"] = f"Bearer {self.token}"
        r.headers.pop("PRIVATE-TOKEN", None)
        r.headers.pop("JOB-TOKEN", None)
        return r


class PrivateTokenAuth(TokenAuth, AuthBase):
    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        r.headers["PRIVATE-TOKEN"] = self.token
        r.headers.pop("JOB-TOKEN", None)
        r.headers.pop("Authorization", None)
        return r


class JobTokenAuth(TokenAuth, AuthBase):
    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        r.headers["JOB-TOKEN"] = self.token
        r.headers.pop("PRIVATE-TOKEN", None)
        r.headers.pop("Authorization", None)
        return r


@dataclasses.dataclass
class SendData:
    content_type: str
    data: dict[str, Any] | MultipartEncoder | None = None
    json: dict[str, Any] | bytes | None = None

    def __post_init__(self) -> None:
        if self.json is not None and self.data is not None:
            raise ValueError(
                f"`json` and `data` are mutually exclusive. Only one can be set. "
                f"json={self.json!r}  data={self.data!r}"
            )


class RequestsResponse(protocol.BackendResponse):
    def __init__(self, response: requests.Response) -> None:
        self._response: requests.Response = response

    @property
    def response(self) -> requests.Response:
        return self._response

    @property
    def status_code(self) -> int:
        return self._response.status_code

    @property
    def headers(self) -> CaseInsensitiveDict[str]:
        return self._response.headers

    @property
    def content(self) -> bytes:
        return self._response.content

    @property
    def reason(self) -> str:
        return self._response.reason

    def json(self) -> Any:
        return self._response.json()


class RequestsBackend(protocol.Backend):
    def __init__(self, session: requests.Session | None = None) -> None:
        self._client: requests.Session = session or requests.Session()

    @property
    def client(self) -> requests.Session:
        return self._client

    @staticmethod
    def prepare_send_data(
        files: dict[str, Any] | None = None,
        post_data: dict[str, Any] | bytes | BinaryIO | None = None,
        raw: bool = False,
    ) -> SendData:
        if files:
            if post_data is None:
                post_data = {}
            else:
                # When creating a `MultipartEncoder` instance with data-types
                # which don't have an `encode` method it will cause an error:
                #       object has no attribute 'encode'
                # So convert common non-string types into strings.
                if TYPE_CHECKING:
                    assert isinstance(post_data, dict)
                for k, v in post_data.items():
                    if isinstance(v, bool):
                        v = int(v)
                    if isinstance(v, (complex, float, int)):
                        post_data[k] = str(v)
            post_data["file"] = files.get("file")
            post_data["avatar"] = files.get("avatar")

            data = MultipartEncoder(fields=post_data)
            return SendData(data=data, content_type=data.content_type)

        if raw and post_data:
            return SendData(data=post_data, content_type="application/octet-stream")

        if TYPE_CHECKING:
            assert not isinstance(post_data, BinaryIO)

        return SendData(json=post_data, content_type="application/json")

    def http_request(
        self,
        method: str,
        url: str,
        json: dict[str, Any] | bytes | None = None,
        data: dict[str, Any] | MultipartEncoder | None = None,
        params: Any | None = None,
        timeout: float | None = None,
        verify: bool | str | None = True,
        stream: bool | None = False,
        **kwargs: Any,
    ) -> RequestsResponse:
        """Make HTTP request

        Args:
            method: The HTTP method to call ('get', 'post', 'put', 'delete', etc.)
            url: The full URL
            data: The data to send to the server in the body of the request
            json: Data to send in the body in json by default
            timeout: The timeout, in seconds, for the request
            verify: Whether SSL certificates should be validated. If
                the value is a string, it is the path to a CA file used for
                certificate validation.
            stream: Whether the data should be streamed

        Returns:
            A requests Response object.
        """
        response: requests.Response = self._client.request(
            method=method,
            url=url,
            params=params,
            data=data,
            timeout=timeout,
            stream=stream,
            verify=verify,
            json=json,
            **kwargs,
        )
        return RequestsResponse(response=response)



from typing import Any

from gitlab.base import RESTObject
from gitlab.mixins import CRUDMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import RequiredOptional

__all__ = ["ProjectMergeRequestDraftNote", "ProjectMergeRequestDraftNoteManager"]


class ProjectMergeRequestDraftNote(ObjectDeleteMixin, SaveMixin, RESTObject):
    def publish(self, **kwargs: Any) -> None:
        path = f"{self.manager.path}/{self.encoded_id}/publish"
        self.manager.gitlab.http_put(path, **kwargs)


class ProjectMergeRequestDraftNoteManager(CRUDMixin[ProjectMergeRequestDraftNote]):
    _path = "/projects/{project_id}/merge_requests/{mr_iid}/draft_notes"
    _obj_cls = ProjectMergeRequestDraftNote
    _from_parent_attrs = {"project_id": "project_id", "mr_iid": "iid"}
    _create_attrs = RequiredOptional(
        required=("note",),
        optional=(
            "commit_id",
            "in_reply_to_discussion_id",
            "position",
            "resolve_discussion",
        ),
    )
    _update_attrs = RequiredOptional(optional=("position",))

    def bulk_publish(self, **kwargs: Any) -> None:
        path = f"{self.path}/bulk_publish"
        self.gitlab.http_post(path, **kwargs)
